/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <config.h>

#ifdef USE_GTK

#include <xapi/wnd.h>
#include <gtk/gtk.h>

int 
api_evt_destroy (GtkWidget* widget, gpointer data)
{
  xl_expr_evaluate (XL_EXPR(data));
  return TRUE;
}

int 
api_evt_button_clicked (GtkWidget* widget, gpointer data)
{
  XLExpr* l_param;

  l_param = xl_expr_get (XL_EXPR(data), 0);
  xl_expr_evaluate (l_param);
  x_access (xl_data_get_raw (l_param->data), 0, xptr) = (xptr) widget;

  xl_expr_evaluate (XL_EXPR(data));

  return TRUE;
}

XLExpr*
xl_expr_create_data_pointer (xptr p_value)
{
  XLExpr* l_expr;
  XLType* l_type;
  XLData* l_data;

  l_type = xl_type_new (XL_TYPE_TYPE_POINTER, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_access (xl_data_get_raw (l_data), 0, xptr) = p_value;
  
  l_expr = xl_expr_new ();
  xl_expr_set_data (l_expr, l_data);

  return l_expr;
}

xvoid
api_wnd_register (xptr p_wnd, XString* p_event, XString* p_callback)
{
  XLExpr* l_callback;

  l_callback = xl_expr_new_fnct (x_string_get_str (p_callback));
  xl_expr_add (l_callback, xl_expr_create_data_pointer (0));

  if (x_string_cmp_str (p_event, "window::close") == 0)
    {
      gtk_signal_connect (GTK_OBJECT (p_wnd), "destroy", 
			  GTK_SIGNAL_FUNC (api_evt_destroy), 
			  (gpointer) l_callback);
    }
  if (x_string_cmp_str (p_event, "button::clicked") == 0)
    {
      gtk_signal_connect (GTK_OBJECT (p_wnd), "clicked", 
			  GTK_SIGNAL_FUNC (api_evt_button_clicked), 
			  (gpointer) l_callback);
    }
}

xvoid
api_wnd_init ()
{
  gtk_init (NULL, NULL);
}

xptr
api_wnd_window_new (XString* p_text, int x, int y, int w, int h)
{
  GtkWidget* window;
  GtkWidget* fixed;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW(window), x_string_get_str (p_text));
  gtk_widget_set_uposition (window, x, y);
  gtk_widget_set_usize (window, w, h);

  fixed = gtk_fixed_new();

  gtk_container_add(GTK_CONTAINER(window), fixed);
  gtk_widget_show (fixed);
  gtk_widget_show (window);

  return (xptr) fixed;
}

xptr
api_wnd_button_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  GtkWidget* button;

  button = gtk_button_new_with_label (x_string_get_str (p_text));
  gtk_widget_set_usize (button, w, h);
  gtk_fixed_put (GTK_FIXED (p_parent), button, x, y);
  gtk_widget_show (button);

  return (xptr) button;
}

xptr
api_wnd_textview_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  return NULL;
}

xptr
api_wnd_label_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  return NULL;
}

xptr
api_wnd_textentry_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  GtkWidget* entry;

  entry = gtk_entry_new ();
  gtk_entry_set_editable (GTK_ENTRY(entry), TRUE);
  gtk_entry_set_visibility (GTK_ENTRY(entry), TRUE);
  gtk_widget_set_usize (entry, w, h);
  gtk_fixed_put (GTK_FIXED (p_parent), entry, x, y);
  gtk_widget_show (entry);

  return (xptr) entry;
}

xvoid
api_wnd_textentry_set_text (xptr p_wnd, XString* p_text)
{
  gtk_entry_set_text (GTK_ENTRY(p_wnd), x_string_get_str (p_text));
}

XString*
api_wnd_textentry_get_text (xptr p_wnd)
{
  XString* l_text;

  l_text = x_string_new ();
  x_string_set_str (l_text, gtk_entry_get_text (GTK_ENTRY(p_wnd)));

  return l_text;
}

xvoid
api_wnd_main ()
{
  gtk_main ();
}

#endif
