/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <config.h>

#ifdef USE_WIN32

#include <xapi/wnd.h>
#include <windows.h>

HFONT	 hFont;

XString* callback = NULL;

XList*   api_wnd_messages = NULL;

XList*   g_api_wnd_event_id_list = NULL;
xu32     g_api_wnd_event_id_count = 1;

xu32 MSG_WND_WINDOW_CLOSE;
xu32 MSG_WND_BUTTON_CLICKED;

typedef struct _APIWndEventId APIWndEventId;
struct _APIWndEventId
{
  XObject  parent;

  XString* event;
  xu32     event_id;
};

xvoid
api_wnd_event_id_destroy (XObject* self)
{
}

APIWndEventId*
api_wnd_event_id_new (xstr p_event)
{
  APIWndEventId* self;
  
  self = x_new (APIWndEventId);
  x_object_init_object (X_OBJECT (self), api_wnd_event_id_destroy);

  self->event = x_string_new ();
  x_string_set_str (self->event, p_event);
  self->event_id = g_api_wnd_event_id_count++;

  return self;
}

xu32
api_wnd_event_id_add (xstr p_event)
{
  APIWndEventId* l_evt;

  l_evt = api_wnd_event_id_new (p_event);
  x_list_add (g_api_wnd_event_id_list, X_OBJECT (l_evt));

  return l_evt->event_id;
}

xu32
api_wnd_event_id_get (XString* p_event)
{
  APIWndEventId* l_evt;
  xu32           i;

  for (i=0; i<x_list_get_size (g_api_wnd_event_id_list); i++)
    {
      l_evt = (APIWndEventId*) x_list_get (g_api_wnd_event_id_list, i);
      if (x_string_cmp (l_evt->event, p_event) == 0)
	return l_evt->event_id;
    }

  return 0;
}

xvoid
api_wnd_event_id_init ()
{
  g_api_wnd_event_id_list = x_list_new ();

  MSG_WND_BUTTON_CLICKED = api_wnd_event_id_add ("button::clicked");
  MSG_WND_WINDOW_CLOSE = api_wnd_event_id_add ("window::close");
}

typedef struct _APIWndEvent APIWndEvent;
struct _APIWndEvent
{
  XObject   parent;

  xptr      wnd;
  xu32      event_id;
  XLExpr*   callback;
};

xvoid
api_wnd_event_destroy (XObject* self)
{
}

APIWndEvent*
api_wnd_event_new (xptr p_wnd, XString* p_event)
{
  APIWndEvent*	self;
  
  self = x_new (APIWndEvent);
  x_object_init_object (X_OBJECT (self), api_wnd_event_destroy);

  self->wnd = p_wnd;
  self->event_id = api_wnd_event_id_get (p_event);

  return self;
}

xvoid
api_wnd_event_raise (xptr p_wnd, xu32 p_event_id)
{
  APIWndEvent*   l_evt;
  XLExpr*        l_param;
  xu32           i;

  for (i=0; i<x_list_get_size (api_wnd_messages); i++)
    {
      l_evt = (APIWndEvent*) x_list_get (api_wnd_messages, i);

      if ((l_evt->event_id == p_event_id) && (l_evt->wnd == p_wnd))
	{
	  l_param = xl_expr_get (l_evt->callback, 0);
	  xl_expr_evaluate (l_param);
	  x_access (xl_data_get_raw (l_param->data), 0, xptr) = p_wnd;

	  xl_expr_evaluate (l_evt->callback);
	}
    }
}

LRESULT CALLBACK api_wnd_wndproc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  APIWndEvent* l_event;
  xu32         i, size;

  if (msg == WM_COMMAND)
    {
      if (HIWORD(wParam) == BN_CLICKED)
	{
	  api_wnd_event_raise ((xptr) lParam, MSG_WND_BUTTON_CLICKED);
	}
    }
  if (msg == WM_CLOSE)
    {
      api_wnd_event_raise ((xptr) hwnd, MSG_WND_WINDOW_CLOSE);
    }
  
  return DefWindowProc (hwnd, msg, wParam, lParam);
}

XLExpr*
xl_expr_create_data_pointer (xptr p_value)
{
  XLExpr* l_expr;
  XLType* l_type;
  XLData* l_data;

  l_type = xl_type_new (XL_TYPE_TYPE_POINTER, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_access (xl_data_get_raw (l_data), 0, xptr) = p_value;
  
  l_expr = xl_expr_new ();
  xl_expr_set_data (l_expr, l_data);

  return l_expr;
}

xvoid
api_wnd_register (xptr p_wnd, XString* p_event, XString* p_callback)
{
  APIWndEvent* l_event;

  l_event = api_wnd_event_new (p_wnd, p_event);
  l_event->callback = xl_expr_new_fnct (x_string_get_str (p_callback));
  xl_expr_add (l_event->callback, xl_expr_create_data_pointer (0));

  x_list_add (api_wnd_messages, X_OBJECT(l_event));
}

xvoid
api_wnd_init ()
{
  WNDCLASSEX	wndclass;
  
  wndclass.cbSize = sizeof(wndclass);
  wndclass.style = CS_HREDRAW | CS_VREDRAW;
  wndclass.lpfnWndProc = api_wnd_wndproc;
  wndclass.cbClsExtra = 0;
  wndclass.cbWndExtra = 0;
  wndclass.hInstance = GetModuleHandle (NULL);
  wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
  wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
  wndclass.hbrBackground = GetSysColorBrush (COLOR_BTNFACE);
  wndclass.lpszMenuName = NULL;
  wndclass.lpszClassName = "XLangWnd";
  wndclass.hIconSm = NULL;
  
  RegisterClassEx (&wndclass);

  hFont = CreateFont (2, 2, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, 
		      DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, 
		      DEFAULT_QUALITY, DEFAULT_PITCH, "MS Sans Serif");

  api_wnd_messages = x_list_new ();

  api_wnd_event_id_init ();

  LoadLibrary ("RICHED32.DLL");
}

xptr
api_wnd_window_new (XString* p_text, int x, int y, int w, int h)
{
  xptr l_wnd;

  l_wnd = CreateWindowEx (0, "XLangWnd", x_string_get_str (p_text), 
			 WS_OVERLAPPEDWINDOW,
			 x, y, w, h, NULL, NULL,
			 GetModuleHandle (NULL), NULL);

  ShowWindow(l_wnd, SW_SHOW); 
  UpdateWindow(l_wnd); 

  return l_wnd;
}

xptr
api_wnd_button_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  xptr l_wnd;

  l_wnd = CreateWindowEx (0, "BUTTON", x_string_get_str (p_text), 
			  WS_VISIBLE | WS_CHILD,
			  x, y, w, h, p_parent, 
			  NULL, GetModuleHandle (NULL), NULL);

  SendMessage (l_wnd, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(1, 0)); 

  return l_wnd;
}

xptr
api_wnd_textview_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  xptr l_wnd;

  l_wnd = CreateWindowEx (0, "RichEdit", x_string_get_str (p_text), 
			  WS_BORDER | ES_MULTILINE | WS_VISIBLE | WS_CHILD,
			  x, y, w, h, p_parent, 
			  NULL, GetModuleHandle (NULL), NULL);

  SendMessage (l_wnd, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(1, 0)); 

  return l_wnd;
}

xptr
api_wnd_label_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  xptr l_wnd;

  l_wnd = CreateWindowEx (0, "STATIC", x_string_get_str (p_text), 
			  WS_VISIBLE | WS_CHILD,
			  x, y, w, h, p_parent, 
			  NULL, GetModuleHandle (NULL), NULL);

  SendMessage (l_wnd, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(1, 0)); 

  return l_wnd;
}

xptr
api_wnd_textentry_new (XString* p_text, int x, int y, int w, int h, xptr p_parent)
{
  xptr l_wnd;

  l_wnd = CreateWindowEx (WS_EX_CLIENTEDGE | WS_EX_RIGHT, "EDIT", x_string_get_str (p_text), 
			  WS_VISIBLE | WS_CHILD,
			  x, y, w, h, p_parent, 
			  NULL, GetModuleHandle (NULL), NULL);

  SendMessage (l_wnd, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(1, 0)); 

  return l_wnd;
}

xvoid
api_wnd_textentry_set_text (xptr p_wnd, XString* p_text)
{
  SetWindowText (p_wnd, x_string_get_str (p_text));
}

XString*
api_wnd_textentry_get_text (xptr p_wnd)
{
  XString* l_text;
  xs8      l_buf[1000];

  GetWindowText (p_wnd, l_buf, 1000);

  l_text = x_string_new ();
  x_string_set_str (l_text, l_buf);

  return l_text;
}

xvoid
api_wnd_main ()
{
  MSG msg;

  while (GetMessage (&msg, NULL, 0, 0))
    {
      TranslateMessage (&msg);
      DispatchMessage (&msg);
    }
}

#endif
