/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

%{

#define yylex  c_lex
#define yylval c_lval
#define yywrap c_wrap

#include <stdio.h>
#include <stdlib.h>

#include <xltype.h>
#include <xlexpr.h>

#include <langc_yacc.h>

extern xu32 c_line;

%}

DIGIT		[0-9]
IDENTIFIER      [a-zA-Z_][a-zA-Z0-9_]*

%x comment

%%

"/*"                 BEGIN(comment);
<comment>[^*\n]*   
<comment>"*"+[^*/\n]*
<comment>\n
<comment>"*"+"/"     BEGIN(INITIAL);

if                   { return IF; }
else                 { return ELSE; }
class                { return CLASS; }
void                 { return VOID; }
char                 { return CHAR; }
uchar                { return UCHAR; }
short                { return SHORT; }
ushort               { return USHORT; }
int                  { return INT; }
uint                 { return UINT; }
float                { return FLOAT; }
double               { return DOUBLE; }
string               { return STRING; }
pointer              { return POINTER; }
handle               { return POINTER; }
bool                 { return BOOL; }
return               { return RETURN; }
while                { return WHILE; }
dynamic              { return DYNAMIC; }
static               { return STATIC; }
true                 { return V_TRUE; }
false                { return V_FALSE; }
{IDENTIFIER}         { 
                       strcpy (yylval.identifier, yytext);
                       return IDENTIFIER; 
                     }
{DIGIT}+             {
                       yylval.val_int = atoi (yytext);
                       return V_INT;
                     }
{DIGIT}+"."{DIGIT}*  {
                       yylval.val_float = atof (yytext);
                       return V_FLOAT;
                     }
\"[^\"]*\"           { 
                       strcpy (yylval.val_string, yytext); 
                       return V_STRING;
                     }
[ \t\r]              /* White spaces */
[\n]                 c_line++;
.		     { return yytext[0]; }






