/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

%{

#include <xltype.h>
#include <xlexpr.h>

#define	yymaxdepth c_maxdepth
#define	yyparse	   c_parse
#define	yylex	   c_lex
#define	yyerror	   c_error
#define	yylval	   c_lval
#define	yychar	   c_char
#define	yydebug	   c_debug
#define	yypact	   c_pact	
#define	yyr1	   c_r1			
#define	yyr2	   c_r2			
#define	yydef	   c_def		
#define	yychk	   c_chk		
#define	yypgo	   c_pgo		
#define	yyact	   c_act		
#define	yyexca	   c_exca
#define yyerrflag  c_errflag
#define yynerrs	   c_nerrs
#define	yyps	   c_ps
#define	yypv	   c_pv
#define	yys	   c_s
#define	yy_yys	   c_yys
#define	yystate	   c_state
#define	yytmp	   c_tmp
#define	yyv	   c_v
#define	yy_yyv	   c_yyv
#define	yyval	   c_val
#define	yylloc	   c_lloc
#define yyreds	   c_reds		/* With YYDEBUG defined */
#define yytoks	   c_toks		/* With YYDEBUG defined */
#define yylhs	   c_yylhs
#define yylen	   c_yylen
#define yydefred   c_yydefred
#define yydgoto	   c_yydgoto
#define yysindex   c_yysindex
#define yyrindex   c_yyrindex
#define yygindex   c_yygindex
#define yytable	   c_yytable
#define yycheck	   c_yycheck

#ifndef YYDEBUG
#define	YYDEBUG	 0	/* Default to no yydebug support */
#endif

#define LANGCDEBUG 1

extern void yyerror (char *error);
extern int yylex () ;

extern XLExpr* g_xl_langc_expr;

XLExpr*  c_call_stack[1000];
xu32     c_call_num = 0;

XLExpr*  c_block_stack[1000];
xu32     c_block_num = 0;

XLExpr*  c_fnct = NULL;
XLExpr*  c_block = NULL;
XLExpr*  c_call = NULL;
XLType*  c_type = NULL;
XLData*  c_data = NULL;
XString* c_text = NULL;

XLExpr*  c_class = NULL;

xu32 c_line = 1;

XLExpr*
xl_lang_type_create (XLTypeType p_type, XLTypeArrayType p_array, xu32 p_size, XString* p_class);

XLExpr*
xl_lang_data_string_create (xstr p_text);

XLExpr*
xl_lang_data_integer_create (xu32 p_int);

XLExpr*
xl_lang_data_float_create (xf32 p_value);

XLExpr*
xl_lang_data_bool_create (xbool p_bool);

xvoid
xl_lang_decl_add (XLExpr* p_expr);

xvoid
xl_lang_class_create (XLExpr* p_identifier);

XLExpr*
xl_lang_class_finish ();

xvoid
xl_lang_class_add (XLExpr* p_expr);

xvoid
xl_lang_method_create (XLExpr* p_type, XLExpr* p_identifier);

xvoid
xl_lang_method_add_param (XLExpr* p_type, XLExpr* p_name);

XLExpr*
xl_lang_method_finish (XLExpr* p_code);

xvoid
xl_lang_fnct_create (XLExpr* p_type, XLExpr* p_identifier);

xvoid
xl_lang_fnct_add_param (XLExpr* p_type, XLExpr* p_name);

XLExpr*
xl_lang_fnct_finish (XLExpr* p_code);

XLExpr*
xl_lang_efnct_create (xstr p_fnct);

XLExpr*
xl_lang_efnct_create1 (xstr p_fnct, XLExpr* p_expr1);

XLExpr*
xl_lang_efnct_create2 (xstr p_fnct, XLExpr* p_expr1, XLExpr* p_expr2);

xvoid
xl_lang_block_begin ();

xvoid
xl_lang_block_add (XLExpr* p_expr);

XLExpr*
xl_lang_block_end ();

xvoid
xl_lang_call_begin (xstr p_identifier);

xvoid
xl_lang_call_add (XLExpr* p_expr);

XLExpr*
xl_lang_call_end ();

%}

%union {
  XLExpr*     expr;
  XLTypeType  xtype;
  double      val_float;
  int         val_int;
  char        val_string[100];
  char        identifier[100];
}

%token IF ELSE CLASS RETURN WHILE DYNAMIC STATIC V_TRUE V_FALSE
%token VOID CHAR UCHAR SHORT USHORT INT UINT FLOAT DOUBLE POINTER STRING BOOL HANDLE
%token<identifier> IDENTIFIER

%token<val_int>    V_INT
%token<val_float>  V_FLOAT
%token<val_string> V_STRING

%type<expr>  decl 
%type<expr>  decl_class decl_global decl_local decl_function
%type<expr>  class_prop class_method class_field
%type<expr>  code_line code_block code_block_or_line code_call code_assign code_return code_while code_if code_if_else
%type<expr>  type expr identifier literal expr_array
%type<xtype> type_basic;

%%

/*************
 * Start
 */

input: 
  /* Nothing */
  | input decl
;

decl:
    decl_class      { xl_lang_decl_add ($1); }
  | decl_global     { xl_lang_decl_add ($1); }
  | decl_function   { xl_lang_decl_add ($1); }
;

/****************
 * Declarations
 */

class_field_list: 
  /* Nothing */
  | class_field_list class_field { xl_lang_class_add ($2); }

class_field:   
    class_prop     { $$ = $1; }
  | class_method   { $$ = $1; }
;

class_prop: 
  type identifier ';'
  { $$ = xl_lang_efnct_create2 ("prop", $1, $2); }
;

class_method: 
  type identifier 
  { xl_lang_method_create ($1, $2);  }
  '(' decl_param_list ')'
  code_block
  { $$ = xl_lang_method_finish ($7); }
;

decl_class: 
  CLASS identifier
  { xl_lang_class_create ($2); }
  '{' class_field_list '}' 
  { $$ = xl_lang_class_finish (); }
;

decl_global:
  type identifier ';'  { $$ = xl_lang_efnct_create2 ("global", $1, $2); }
;

decl_local: 
  type identifier ';'  { $$ = xl_lang_efnct_create2 ("local", $1, $2); }
;

decl_param:
  type identifier      { xl_lang_fnct_add_param ($1, $2); }
;

decl_param_list: 
  /* Nothing */
  | decl_param
  | decl_param_list ',' decl_param
;

decl_function: 
  type identifier
  { xl_lang_fnct_create ($1, $2);  }
  '(' decl_param_list ')' 
  code_block
  { $$ = xl_lang_fnct_finish ($7); }
;

/*****************
 * Basic elements
 */

type_basic:
    VOID      { $$ = XL_TYPE_TYPE_VOID; }
  | CHAR      { $$ = XL_TYPE_TYPE_S8; }
  | UCHAR     { $$ = XL_TYPE_TYPE_U8; }
  | SHORT     { $$ = XL_TYPE_TYPE_S32; }
  | USHORT    { $$ = XL_TYPE_TYPE_U16; }
  | INT       { $$ = XL_TYPE_TYPE_S32; }
  | UINT      { $$ = XL_TYPE_TYPE_U32; }
  | STRING    { $$ = XL_TYPE_TYPE_STRING; }
  | POINTER   { $$ = XL_TYPE_TYPE_POINTER; }
  | FLOAT     { $$ = XL_TYPE_TYPE_F32; }
  | DOUBLE    { $$ = XL_TYPE_TYPE_F64; }
;

type:   
    type_basic '[' V_INT ']'   { $$ = xl_lang_type_create ($1, XL_TYPE_ARRAY_TYPE_FIXED, $3, NULL); }
  | type_basic '[' ']'         { $$ = xl_lang_type_create ($1, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL); }
  | type_basic                 { $$ = xl_lang_type_create ($1, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL); } 
  | DYNAMIC '<' identifier '>' { $$ = xl_expr_new (); }
  | STATIC '<' identifier '>'  { $$ = xl_expr_new (); } 
;

identifier:
  IDENTIFIER 
  { $$ =  xl_expr_new_identifier ($1); }
;

literal:   
    V_STRING   { $$ = xl_lang_data_string_create ($1);   }
  | V_INT      { $$ = xl_lang_data_integer_create ($1);  }
  | V_FLOAT    { $$ = xl_lang_data_float_create ($1);    }
  | V_TRUE     { $$ = xl_lang_data_bool_create (TRUE);   }
  | V_FALSE    { $$ = xl_lang_data_bool_create (FALSE);  }
;

/*******************
 * Expressions
 */

expr:   
    literal            { $$ = $1; }
  | code_call          { $$ = $1; }
  | expr_array         { $$ = $1; }
  | identifier         { $$ = $1; } 
  | expr '+' expr      { $$ = xl_lang_efnct_create2 ("+", $1, $3); }
  | expr '*' expr      { $$ = xl_lang_efnct_create2 ("*", $1, $3); }
  | expr '-' expr      { $$ = xl_lang_efnct_create2 ("-", $1, $3); }
  | expr '/' expr      { $$ = xl_lang_efnct_create2 ("/", $1, $3); }
  | expr '<' '=' expr  { $$ = xl_lang_efnct_create2 ("<=", $1, $4); }
  | expr '>' '=' expr  { $$ = xl_lang_efnct_create2 (">=", $1, $4); }
  | expr '<' expr      { $$ = xl_lang_efnct_create2 ("<", $1, $3); }
  | expr '>' expr      { $$ = xl_lang_efnct_create2 (">", $1, $3); }
  | expr '=' '=' expr  { $$ = xl_lang_efnct_create2 ("==", $1, $4); }
  | expr '<' '>' expr  { $$ = xl_lang_efnct_create2 ("<>", $1, $4); }
;

expr_array:
  expr '[' expr ']' { $$ = xl_lang_efnct_create2 ("[]", $1, $3); }
;

param_list: 
  /* Nothing */
  | expr                 { xl_lang_call_add ($1); }
  | param_list ',' expr  { xl_lang_call_add ($3); }
;

/*********************
 * Statements (CODE)
 */


code_lines: 
  /* Nothing */
  | code_lines code_line 
;

code_line:   
    decl_local     { xl_lang_block_add ($1); }
  | code_block     { xl_lang_block_add ($1); }
  | code_call ';'  { xl_lang_block_add ($1); }
  | code_assign    { xl_lang_block_add ($1); }
  | code_return    { xl_lang_block_add ($1); }
  | code_while     { xl_lang_block_add ($1); }
  | code_if        { xl_lang_block_add ($1); }
;

code_block: 
  '{' 
  { xl_lang_block_begin (); }  
  code_lines 
  '}'
  { $$ = xl_lang_block_end (); }
;

code_block_or_line:
    code_block { $$ = $1; }
    | { xl_lang_block_begin (); } code_line { $$ = xl_lang_block_end (); }
;

code_call: 
  identifier
  { xl_lang_call_begin (x_string_get_str ($1->identifier)); }
  '(' param_list ')'
  { $$ = xl_lang_call_end (); }
;

code_assign: 
  expr '=' expr ';' { $$ = xl_lang_efnct_create2 ("=", $1, $3); }
;

code_return: 
  RETURN expr ';'   { $$ = xl_lang_efnct_create1 ("return", $2); }
;

code_while:  
  WHILE '(' expr ')' code_block
  { $$ = xl_lang_efnct_create2 ("while", $3, $5); }
;

code_if_else:
  /* empty */ { $$ = NULL; }
  | ELSE code_block_or_line { $$ = $2; }
;

code_if:  
  IF 
  '(' expr ')' code_block_or_line
  code_if_else
  { if ($5)
      $$ = xl_lang_efnct_create2 ("if", $3, $5); 
    else
      $$ = xl_lang_efnct_create1 ("if", $3);
  }
;

%%

XLExpr*
xl_lang_type_create (XLTypeType p_type, XLTypeArrayType p_array, xu32 p_size, XString* p_class)
{
  XLExpr* l_expr;
  XLType* l_type;

  l_type = xl_type_new (p_type, p_array, p_size, p_class);
  l_expr = xl_expr_new ();
  xl_expr_set_type (l_expr, l_type);

  return l_expr;
}

xvoid
xl_lang_decl_add (XLExpr* p_expr)
{
  xl_expr_add (g_xl_langc_expr, p_expr);
}

xvoid
xl_lang_class_create (XLExpr* p_identifier)
{
  c_class = xl_expr_new_fnct ("class");
  xl_expr_add (c_class, p_identifier);
  xl_expr_add (c_class, xl_expr_new_identifier (""));
}

XLExpr*
xl_lang_class_finish ()
{
  return c_class;
}

xvoid
xl_lang_class_add (XLExpr* p_expr)
{
}

XLExpr*
xl_lang_data_string_create (xstr p_text)
{
  XLExpr*  l_expr;
  XLType*  l_type;
  XLData*  l_data;
  XString* l_text;
  xu32     i;

  l_type = xl_type_new (XL_TYPE_TYPE_STRING, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  l_text = x_string_new ();

  /* Building STRING
   */
  for (i=1; i<strlen(p_text)-1; i++)
    {
      if (p_text[i] == '\\')
	{
	  i++;
	  switch (p_text[i])
	    {
	    case 'n':
	      x_string_add_char (l_text, '\n');
	      break;
	    case 'r':
	      x_string_add_char (l_text, '\r');
	      break;
	    case 't':
	      x_string_add_char (l_text, '\t');
	      break;
	    default:
	      x_string_add_char (l_text, p_text[i]);
	      break;
	    }
	}
      else
	x_string_add_char (l_text, p_text[i]);
    }

  xl_data_set_string (l_data, l_text);
  
  l_expr = xl_expr_new ();
  xl_expr_set_data (l_expr, l_data);

  return l_expr;
}

XLExpr*
xl_lang_data_integer_create (xu32 p_int)
{
  XLExpr* l_expr;
  XLType* l_type;
  XLData* l_data;

  l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_access (xl_data_get_raw (l_data), 0, xu32) = p_int;
  
  l_expr = xl_expr_new ();
  xl_expr_set_data (l_expr, l_data);

  return l_expr;
}

XLExpr*
xl_lang_data_float_create (xf32 p_value)
{
  XLExpr* l_expr;
  XLType* l_type;
  XLData* l_data;

  l_type = xl_type_new (XL_TYPE_TYPE_F32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_access (xl_data_get_raw (l_data), 0, xf32) = p_value;
  
  l_expr = xl_expr_new ();
  xl_expr_set_data (l_expr, l_data);

  return l_expr;
}

XLExpr*
xl_lang_data_bool_create (xbool p_bool)
{
  XLExpr* l_expr;
  XLType* l_type;
  XLData* l_data;

  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_access (xl_data_get_raw (l_data), 0, xbool) = p_bool;
  
  l_expr = xl_expr_new ();
  xl_expr_set_data (l_expr, l_data);

  return l_expr;
}

XLExpr*
xl_lang_efnct_create (xstr p_fnct)
{
  XLExpr* l_expr;

  l_expr = xl_expr_new_fnct (p_fnct);

  return l_expr;
}

XLExpr*
xl_lang_efnct_create1 (xstr p_fnct, XLExpr* p_expr1)
{
  XLExpr* l_expr;

  l_expr = xl_expr_new_fnct (p_fnct);
  xl_expr_add (l_expr, p_expr1);

  return l_expr;
}

XLExpr*
xl_lang_efnct_create2 (xstr p_fnct, XLExpr* p_expr1, XLExpr* p_expr2)
{
  XLExpr* l_expr;

  l_expr = xl_expr_new_fnct (p_fnct);
  xl_expr_add (l_expr, p_expr1);
  xl_expr_add (l_expr, p_expr2);

  return l_expr;
}

xvoid
xl_lang_fnct_create (XLExpr* p_type, XLExpr* p_identifier)
{
  c_fnct = xl_expr_new_fnct ("fnct");
  xl_expr_add (c_fnct, p_type);
  xl_expr_add (c_fnct, p_identifier);
}

xvoid
xl_lang_fnct_add_param (XLExpr* p_type, XLExpr* p_name)
{
  xl_expr_add (c_fnct, p_type);
  xl_expr_add (c_fnct, p_name);
}

XLExpr*
xl_lang_fnct_finish (XLExpr* p_code)
{
  XLExpr* l_fnct;

  xl_expr_add (c_fnct, p_code);
  l_fnct = c_fnct;
  c_fnct = NULL;

  return l_fnct;
}

xvoid
xl_lang_method_create (XLExpr* p_type, XLExpr* p_identifier)
{
  c_fnct = xl_expr_new_fnct ("meth");
  xl_expr_add (c_fnct, p_type);
  xl_expr_add (c_fnct, p_identifier);
}

XLExpr*
xl_lang_method_finish (XLExpr* p_code)
{
  XLExpr* l_fnct;

  xl_expr_add (c_fnct, p_code);
  l_fnct = c_fnct;
  c_fnct = NULL;

  return l_fnct;
}

xvoid
xl_lang_block_begin ()
{
  if (c_block)
    {
      c_block_stack[c_block_num] = c_block;
      c_block_num++;
    }
  c_block = xl_expr_new_fnct ("@");
}

xvoid
xl_lang_block_add (XLExpr* p_expr)
{
  xl_expr_add (c_block, p_expr);
}

XLExpr*
xl_lang_block_end ()
{
  XLExpr* l_block;

  if (c_block_num > 0)
    {
      l_block = c_block;
      c_block_num--;
      c_block = c_block_stack[c_block_num];
    }
  else
    {
      l_block = c_block;
      c_block = NULL;
    }

  return l_block;
}

xvoid
xl_lang_call_begin (xstr p_identifier)
{
  if (c_call)
    {
      c_call_stack[c_call_num] = c_call;
      c_call_num++;
    }
  c_call = xl_expr_new_fnct (p_identifier);
}

xvoid
xl_lang_call_add (XLExpr* p_expr)
{
  xl_expr_add (c_call, p_expr);
}

XLExpr*
xl_lang_call_end ()
{
  XLExpr* l_call;

  if (c_call_num > 0)
    {
      l_call = c_call;
      c_call_num--;
      c_call = c_call_stack[c_call_num];
    }
  else
    {
      l_call = c_call;
      c_call = NULL;
    }

  return l_call;
}
