/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <config.h>
#include <xlapi.h>
#include <xapi/lang.h>
#include <xapi/sys.h>
#include <xapi/io.h>
#include <xapi/wnd.h>

#include <xlmain.h>

XLFnct*
xl_api_add_fnct (xstr p_name, xptr p_fnct, XLTypeType p_type)
{
  XLFnct* l_fnct;

  l_fnct = xl_main_add_fnct_native (g_xl_main_global, p_name, p_fnct);
  l_fnct->ret_type = xl_type_new (p_type, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);  

  return l_fnct;
}

xvoid
xl_api_add_param (XLFnct* l_fnct, xstr p_name, XLTypeType p_type)
{
  XLField* l_field;
  XString* l_name;
  XLType*  l_type;

  l_name = x_string_new ();
  x_string_set_str (l_name, p_name);

  l_type = xl_type_new (p_type, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_field = xl_field_new (l_name, l_type);
  xl_fnct_add_param (l_fnct, l_field);
}

xvoid
xl_api_init ()
{
  XLFnct* l_fnct;

  /* LANG library registration
   */

  l_fnct = xl_api_add_fnct ("lang_itostr", api_lang_itostr, XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "i", XL_TYPE_TYPE_U32);

  l_fnct = xl_api_add_fnct ("lang_strtoi", api_lang_strtoi, XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "str", XL_TYPE_TYPE_STRING);

  l_fnct = xl_api_add_fnct ("lang_ftostr", api_lang_ftostr, XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "i", XL_TYPE_TYPE_F32);

  l_fnct = xl_api_add_fnct ("lang_strtof", api_lang_strtof, XL_TYPE_TYPE_F32);
  xl_api_add_param (l_fnct, "str", XL_TYPE_TYPE_STRING);

  l_fnct = xl_api_add_fnct ("lang_stradd", api_lang_stradd, XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "str", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "stradd", XL_TYPE_TYPE_STRING);

  /* SYS library registration
   */
  
  l_fnct = xl_api_add_fnct ("sys_exit", api_sys_exit, XL_TYPE_TYPE_VOID);
  xl_api_add_param (l_fnct, "code", XL_TYPE_TYPE_U32);

  /* IO library registration
   */
  
  l_fnct = xl_api_add_fnct ("io_write", api_io_write, XL_TYPE_TYPE_VOID);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);

  /* WND library registration
   */

#ifdef USE_GUI

  api_wnd_init ();

  l_fnct = xl_api_add_fnct ("wnd_window_new", api_wnd_window_new, XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "x", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "y", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "w", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "h", XL_TYPE_TYPE_U32); 

  l_fnct = xl_api_add_fnct ("wnd_register", api_wnd_register, XL_TYPE_TYPE_VOID);
  xl_api_add_param (l_fnct, "wnd", XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "event", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "callback", XL_TYPE_TYPE_STRING);

  l_fnct = xl_api_add_fnct ("wnd_button_new", api_wnd_button_new, XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "x", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "y", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "w", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "h", XL_TYPE_TYPE_U32); 
  xl_api_add_param (l_fnct, "parent", XL_TYPE_TYPE_POINTER); 

  l_fnct = xl_api_add_fnct ("wnd_textview_new", api_wnd_textview_new, XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "x", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "y", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "w", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "h", XL_TYPE_TYPE_U32); 
  xl_api_add_param (l_fnct, "parent", XL_TYPE_TYPE_POINTER);

  l_fnct = xl_api_add_fnct ("wnd_label_new", api_wnd_label_new, XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "x", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "y", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "w", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "h", XL_TYPE_TYPE_U32); 
  xl_api_add_param (l_fnct, "parent", XL_TYPE_TYPE_POINTER); 

  l_fnct = xl_api_add_fnct ("wnd_textentry_new", api_wnd_textentry_new, XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "x", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "y", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "w", XL_TYPE_TYPE_U32);
  xl_api_add_param (l_fnct, "h", XL_TYPE_TYPE_U32); 
  xl_api_add_param (l_fnct, "parent", XL_TYPE_TYPE_POINTER); 

  l_fnct = xl_api_add_fnct ("wnd_textentry_get_text", api_wnd_textentry_get_text, XL_TYPE_TYPE_STRING);
  xl_api_add_param (l_fnct, "wnd", XL_TYPE_TYPE_POINTER);

  l_fnct = xl_api_add_fnct ("wnd_textentry_set_text", api_wnd_textentry_set_text, XL_TYPE_TYPE_VOID);
  xl_api_add_param (l_fnct, "wnd", XL_TYPE_TYPE_POINTER);
  xl_api_add_param (l_fnct, "text", XL_TYPE_TYPE_STRING);

  l_fnct = xl_api_add_fnct ("wnd_main", api_wnd_main, XL_TYPE_TYPE_VOID);

#endif
}

xvoid
xl_api_exit ()
{
}

