/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <xlbuiltin.h>
#include <xllang.h>
#include <xlclass.h>
#include <xlmain.h>

XLClass* g_xl_builtin_current_class = NULL;

#define XL_ERROR printf
#define XL_ERR_BAD_EXPR   "ERROR('%s'): '%s' expression of bad type\n"

xvoid
xl_builtin_init ()
{
  /* Declarators
   */

  xl_main_add_fnct_builtin (g_xl_main_global, "fnct", xl_builtin_fnct);
  xl_main_add_fnct_builtin (g_xl_main_global, "global", xl_builtin_global);
  xl_main_add_fnct_builtin (g_xl_main_global, "local", xl_builtin_local);
  xl_main_add_fnct_builtin (g_xl_main_global, "class", xl_builtin_class); 
  xl_main_add_fnct_builtin (g_xl_main_global, "meth", xl_builtin_meth);
  xl_main_add_fnct_builtin (g_xl_main_global, "prop", xl_builtin_prop);

  xl_main_add_fnct_builtin (g_xl_main_global, "+", xl_builtin_add);
  xl_main_add_fnct_builtin (g_xl_main_global, "-", xl_builtin_sub);
  xl_main_add_fnct_builtin (g_xl_main_global, "*", xl_builtin_mul);
  xl_main_add_fnct_builtin (g_xl_main_global, "/", xl_builtin_div);

  xl_main_add_fnct_builtin (g_xl_main_global, "import", xl_builtin_import);

  xl_main_add_fnct_builtin (g_xl_main_global, "new", xl_builtin_new);  
  xl_main_add_fnct_builtin (g_xl_main_global, ".", xl_builtin_access);  
  xl_main_add_fnct_builtin (g_xl_main_global, "::", xl_builtin_invoke);
  xl_main_add_fnct_builtin (g_xl_main_global, "[]", xl_builtin_array);

  xl_main_add_fnct_builtin (g_xl_main_global, "return", xl_builtin_return);
  xl_main_add_fnct_builtin (g_xl_main_global, "if", xl_builtin_if);
  xl_main_add_fnct_builtin (g_xl_main_global, "while", xl_builtin_while);

  xl_main_add_fnct_builtin (g_xl_main_global, "<", xl_builtin_cmp_l);
  xl_main_add_fnct_builtin (g_xl_main_global, ">", xl_builtin_cmp_g);
  xl_main_add_fnct_builtin (g_xl_main_global, "<=", xl_builtin_cmp_le);
  xl_main_add_fnct_builtin (g_xl_main_global, ">=", xl_builtin_cmp_ge);
  xl_main_add_fnct_builtin (g_xl_main_global, "<>", xl_builtin_cmp_ne);
  xl_main_add_fnct_builtin (g_xl_main_global, "==", xl_builtin_cmp_e);

  xl_main_add_fnct_builtin (g_xl_main_global, "=", xl_builtin_assign);

  xl_main_add_fnct_builtin (g_xl_main_global, "++", xl_builtin_inc);
  xl_main_add_fnct_builtin (g_xl_main_global, "--", xl_builtin_dec);
  
  xl_main_add_fnct_builtin (g_xl_main_global, "@", xl_builtin_block);
}

xvoid
xl_builtin_exit ()
{
}

xvoid
xl_builtin_fnct (XLExpr* p_expr)
{
  XLFnct*  l_fnct;
  XLExpr*  l_expr;
  XLType*  l_return_type;
  XString* l_function_name;
  XLType*  l_param_type;
  XString* l_param_name;
  XLField* l_field;
  XLExpr*  l_code;
  xu32     l_size;
  xu32     i;

  l_size = xl_expr_get_size (p_expr);


  l_expr = xl_expr_get (p_expr, 0);
  if (xl_expr_check (l_expr, XL_EXPR_TYPE_TYPE))
    {
      l_return_type = x_addref (XLType, l_expr->type);
      x_unref (l_expr);
    }
  else
    {
      XL_ERROR (XL_ERR_BAD_EXPR, "fnct", "return-type");
      x_unref (l_expr);
      return;
    }

  l_expr = xl_expr_get (p_expr, 1);
  if (xl_expr_check (l_expr, XL_EXPR_TYPE_IDENTIFIER))
    {
      l_function_name = x_addref (XString, l_expr->identifier);
      x_unref (l_expr);
    }
  else
    {
      XL_ERROR (XL_ERR_BAD_EXPR, "fnct", "function-name");
      x_unref (l_expr);
      return;
    }  

  l_code = xl_expr_get (p_expr, l_size - 1);
  l_fnct = xl_main_add_fnct_expr (g_xl_main_global, x_string_get_str (l_function_name), l_code);
  x_unref (l_code);

  for (i = 0; i < ((l_size - 3) / 2); i++)
    {
      l_expr = xl_expr_get (p_expr, 2 + (i * 2));
      if (xl_expr_check (l_expr, XL_EXPR_TYPE_TYPE))
	{
	  l_param_type = x_addref (XLType, l_expr->type);
	  x_unref (l_expr);
	}
      else
	{
	  XL_ERROR (XL_ERR_BAD_EXPR, "fnct", "param-type");
	  x_unref (l_expr);
	  return;
	}

      l_expr = xl_expr_get (p_expr, 2 + (i * 2) + 1);
      if (xl_expr_check (l_expr, XL_EXPR_TYPE_IDENTIFIER))
	{
	  l_param_name = x_addref (XString, l_expr->identifier);
	  x_unref (l_expr);
	}
      else
	{
	  XL_ERROR (XL_ERR_BAD_EXPR, "fnct", "param-type");
	  x_unref (l_expr);
	  return;
	}

      l_field = xl_field_new (l_param_name, l_param_type);
      xl_fnct_add_param (l_fnct, l_field);

      x_unref (l_param_type);
      x_unref (l_param_name); 
      x_unref (l_field);
    }
}

xvoid
xl_builtin_local (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XLType*  l_type;
  XString* l_name;
  XLData*  l_data;
  XLVar*   l_var;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_type = l_expr->type;
  x_unref (l_expr);

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  l_name = l_expr->identifier;
  x_unref (l_expr);

  l_data = xl_data_new (l_type, NULL);
  l_var = xl_var_new (l_name, l_data);

  xl_main_add_var (g_xl_main_global, l_var, FALSE);

  x_unref (l_data);
  x_unref (l_var);
}

xvoid
xl_builtin_global (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XLType*  l_type;
  XString* l_name;
  XLData*  l_data;
  XLVar*   l_var;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_type = l_expr->type;
  x_unref (l_expr);

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  l_name = l_expr->identifier;
  x_unref (l_expr);

  l_data = xl_data_new (l_type, NULL);
  l_var = xl_var_new (l_name, l_data);

  xl_main_add_var (g_xl_main_global, l_var, TRUE);

  x_unref (l_data);
  x_unref (l_var);
}

xvoid
xl_builtin_class (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XString* l_text;
  XLClass* l_class;
  xu32     l_size, i;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_text = l_expr->identifier;

  l_class = xl_class_new (l_text);
  g_xl_builtin_current_class = l_class;
  l_size = x_list_get_size (p_expr->list_expr);

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  if (l_expr->expr_type == XL_EXPR_TYPE_IDENTIFIER)
    i = 2;
  else
    i = 1;

  for (; i<l_size; i++)
    {
      l_expr = xl_expr_get (p_expr, i);
      xl_expr_evaluate (l_expr);
      x_unref (l_expr);
    }
  
  g_xl_builtin_current_class = NULL;
  xl_main_add_class (g_xl_main_global, l_class);
}

xvoid
xl_builtin_prop (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XLType*  l_prop_type;
  XString* l_prop_name;
  XLField* l_field;
  xu32     l_size;


  l_size = xl_expr_get_size (p_expr);

  l_expr = xl_expr_get (p_expr, 0);
  if (xl_expr_check (l_expr, XL_EXPR_TYPE_TYPE))
    {
      l_prop_type = x_addref (XLType, l_expr->type);
      x_unref (l_expr);
    }
  else
    {
      XL_ERROR (XL_ERR_BAD_EXPR, "prop", "prop-type");
      x_unref (l_expr);
      return;
    }

  l_expr = xl_expr_get (p_expr, 1);
  if (xl_expr_check (l_expr, XL_EXPR_TYPE_IDENTIFIER))
    {
      l_prop_name = x_addref (XString, l_expr->identifier);
      x_unref (l_expr);
    }
  else
    {
      XL_ERROR (XL_ERR_BAD_EXPR, "prop", "prop-name");
      x_unref (l_expr);
      return;
    }  

  l_field = xl_field_new (l_prop_name, l_prop_type);
  xl_class_add_field (g_xl_builtin_current_class, l_field);

  x_unref (l_field);
  x_unref (l_prop_type);
  x_unref (l_prop_name);
}

xvoid
xl_builtin_meth (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XLExpr*  l_subexpr;
  XString* l_name;
  XLFnct*  l_fnct;
  XLField* l_field;
  xu32	   l_size;
  xu32     i;

  l_size = x_list_get_size (p_expr->list_expr);
  
  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  l_name = x_string_new ();
  x_string_set (l_name, g_xl_builtin_current_class->name);
  x_string_add_str (l_name, "::");
  x_string_add (l_name, l_expr->identifier);

  x_unref (l_expr);

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, l_size - 1));

  l_fnct = xl_main_add_fnct_expr (g_xl_main_global, x_string_get_str (l_name), l_expr);
  x_unref (l_expr);

  for (i=0; i<(l_size / 2) - 1; i++)
    {
      l_subexpr = XL_EXPR (x_list_get (p_expr->list_expr, (i * 2) + 3));

      l_field = xl_field_new (l_subexpr->identifier, NULL);
      xl_fnct_add_param (l_fnct, l_field);

      x_unref (l_field);
      x_unref (l_subexpr);
    }
}

xvoid
xl_builtin_access (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XLData*  l_data;
  XString* l_field;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);

  l_data = l_expr->data;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  l_field = l_expr->identifier;

  p_expr->data = xl_data_get_field (l_data, l_field);
}

xvoid
xl_builtin_invoke (XLExpr* p_expr)
{
  XLExpr*  l_expr;
  XLExpr*  l_subexpr;
  XLData*  l_data;
  XString* l_name;
  xu32     l_size, i;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);
  l_data = l_expr->data;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  l_name = x_string_new ();
  x_string_set (l_name, l_data->type->class->name);
  x_string_add_str (l_name, "::");
  x_string_add (l_name, l_expr->identifier);

  l_expr = xl_expr_new ();
  xl_expr_set_fnct (l_expr, l_name);

  l_size = x_list_get_size (p_expr->list_expr);
  for (i=2; i<l_size; i++)
    {
      l_subexpr = XL_EXPR (x_list_get (p_expr->list_expr, i));
      xl_expr_add (l_expr, l_subexpr);
    }

  xl_expr_evaluate (l_expr);
}

xvoid
xl_builtin_array (XLExpr* p_expr)
{
  XLExpr* l_expr;
  XLData* l_data;
  XLData* l_index;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);
  l_data = x_addref (XLData, l_expr->data);
  x_unref (l_expr);

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  xl_expr_evaluate (l_expr);
  l_index = x_addref (XLData, l_expr->data);
  x_unref (l_expr);

  p_expr->data = xl_data_get_index (l_data, x_access (xl_data_get_raw (l_index), 0, xu32));

  x_unref (l_data);
  x_unref (l_index);
}

xvoid
xl_builtin_new (XLExpr* p_expr)
{
  printf ("new\n");
}

xvoid
xl_builtin_assign  (XLExpr* p_expr)
{
  XLExpr* l_expr;
  XLData* l_src;
  XLData* l_dst;

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);
  l_dst = x_addref (XLData, l_expr->data);
  x_unref (l_expr);

  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 1));  
  xl_expr_evaluate (l_expr);
  l_src = x_addref (XLData, l_expr->data);
  x_unref (l_expr);

  xl_data_assign (l_dst, l_src);

  x_unref (l_dst);
  x_unref (l_src);
}

xvoid
xl_builtin_return (XLExpr* p_expr)
{
  XLExpr* l_expr;
  
  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);

  g_xl_fnct_return = x_addref (XLData, l_expr->data);

  x_unref (l_expr);
}

xvoid
xl_builtin_if (XLExpr* p_expr)
{
  XLExpr* l_test;
  XLExpr* l_true;
  XLExpr* l_false;
  
  l_test = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_true = XL_EXPR (x_list_get (p_expr->list_expr, 1));
  l_false = XL_EXPR (x_list_get (p_expr->list_expr, 2));

  xl_expr_evaluate (l_test);

  if (x_access (xl_data_get_raw (l_test->data), 0, xbool))
    {
      xl_expr_evaluate (l_true);
    }
  else
    {
      if (l_false)
	xl_expr_evaluate (l_false);
    }

  x_unref (l_test);
  x_unref (l_true);
  x_unref (l_false);
}

xvoid
xl_builtin_while (XLExpr* p_expr)
{
  XLExpr* l_test;
  XLExpr* l_code;
  
  l_test = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_code = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_test);
  while (x_access (xl_data_get_raw (l_test->data), 0, xbool))
    {
      xl_expr_evaluate (l_code);
      xl_expr_evaluate (l_test);
    }

  x_unref (l_test);
  x_unref (l_code);
}

xvoid
xl_builtin_inc (XLExpr* p_expr)
{
  XLExpr* l_expr;
  
  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);
  x_unref (l_expr);

  x_access (xl_data_get_raw (l_expr->data), 0, xu32) = x_access (xl_data_get_raw (l_expr->data), 0, xu32) + 1;
}

xvoid
xl_builtin_dec (XLExpr* p_expr)
{
  XLExpr* l_expr;
  
  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);
  x_unref (l_expr);

  x_access (xl_data_get_raw (l_expr->data), 0, xu32) = x_access (xl_data_get_raw (l_expr->data), 0, xu32) - 1;
}

xvoid
xl_builtin_cmp_l (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);
 
  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_unref (l_type);

  if (x_access (xl_data_get_raw (l_left->data), 0, xu32) < x_access (xl_data_get_raw (l_right->data), 0, xu32))
    x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
  else
    x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_cmp_g (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);
 
  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_unref (l_type);

  if (x_access (xl_data_get_raw (l_left->data), 0, xu32) > x_access (xl_data_get_raw (l_right->data), 0, xu32))
    x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
  else
    x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_cmp_le (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);
 
  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_unref (l_type);

  if (x_access (xl_data_get_raw (l_left->data), 0, xu32) <= x_access (xl_data_get_raw (l_right->data), 0, xu32))
    x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
  else
    x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_cmp_ge (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);
 
  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_unref (l_type);

  if (x_access (xl_data_get_raw (l_left->data), 0, xu32) >= x_access (xl_data_get_raw (l_right->data), 0, xu32))
    x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
  else
    x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_cmp_ne (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);
 
  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_unref (l_type);

  if (x_access (xl_data_get_raw (l_left->data), 0, xu32) != x_access (xl_data_get_raw (l_right->data), 0, xu32))
    x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
  else
    x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_cmp_e (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);
 
  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, NULL);
  x_unref (l_type);

  if (x_access (xl_data_get_raw (l_left->data), 0, xu32) == x_access (xl_data_get_raw (l_right->data), 0, xu32))
    x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
  else
    x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_io_write (XLExpr* p_expr)
{
  XLExpr* l_expr;
  XLData* l_data;
  XString*l_text;
  
  l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  xl_expr_evaluate (l_expr);

  l_data = l_expr->data;
  x_unref (l_expr);

  if (!l_data)
    {
      printf ("ERROR: data = NULL\n");
      return;
    }
  
  if (l_data->type->type == XL_TYPE_TYPE_STRING)
    {
      l_text = xl_data_get_string (l_data);
      printf ("%s", x_string_get_str (l_text));
      x_unref (l_text);
    }

  if (l_data->type->type == XL_TYPE_TYPE_U32)
    {
      printf ("%d", x_access (xl_data_get_raw (l_data), 0, xu32));
    }
}

xvoid
xl_builtin_block (XLExpr* p_expr)
{
  XLExpr*	 l_expr;
  xu32		 l_size;
  xu32		 i;
  
  l_size = x_list_get_size (p_expr->list_expr);
  
  for (i = 0; i < l_size; i++)
    {
      l_expr = XL_EXPR (x_list_get (p_expr->list_expr, i));
      xl_expr_evaluate (l_expr);
      x_unref (l_expr);
    }
}

xvoid
xl_builtin_add (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);

  if (l_left->data->type->type == XL_TYPE_TYPE_U32)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);
      
      x_access (xl_data_get_raw (l_data), 0, xu32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xu32) + 
	x_access (xl_data_get_raw (l_right->data), 0, xu32);
    }

  if (l_left->data->type->type == XL_TYPE_TYPE_F32)
    {

      l_type = xl_type_new (XL_TYPE_TYPE_F32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);

      x_access (xl_data_get_raw (l_data), 0, xf32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xf32) + 
	x_access (xl_data_get_raw (l_right->data), 0, xf32);
    }

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_sub (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);

  if (l_left->data->type->type == XL_TYPE_TYPE_U32)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);
      
      x_access (xl_data_get_raw (l_data), 0, xu32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xu32) - 
	x_access (xl_data_get_raw (l_right->data), 0, xu32);
    }

  if (l_left->data->type->type == XL_TYPE_TYPE_F32)
    {

      l_type = xl_type_new (XL_TYPE_TYPE_F32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);

      x_access (xl_data_get_raw (l_data), 0, xf32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xf32) - 
	x_access (xl_data_get_raw (l_right->data), 0, xf32);
    }

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_mul (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);

  if (l_left->data->type->type == XL_TYPE_TYPE_U32)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);
      
      x_access (xl_data_get_raw (l_data), 0, xu32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xu32) * 
	x_access (xl_data_get_raw (l_right->data), 0, xu32);
    }

  if (l_left->data->type->type == XL_TYPE_TYPE_F32)
    {

      l_type = xl_type_new (XL_TYPE_TYPE_F32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);

      x_access (xl_data_get_raw (l_data), 0, xf32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xf32) *
	x_access (xl_data_get_raw (l_right->data), 0, xf32);
    }

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_div (XLExpr* p_expr)
{
  XLExpr* l_left;
  XLExpr* l_right;
  XLType* l_type;
  XLData* l_data;

  l_left = XL_EXPR (x_list_get (p_expr->list_expr, 0));
  l_right = XL_EXPR (x_list_get (p_expr->list_expr, 1));

  xl_expr_evaluate (l_left);
  xl_expr_evaluate (l_right);

  if (l_left->data->type->type == XL_TYPE_TYPE_U32)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);
      
      x_access (xl_data_get_raw (l_data), 0, xu32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xu32) / 
	x_access (xl_data_get_raw (l_right->data), 0, xu32);
    }

  if (l_left->data->type->type == XL_TYPE_TYPE_F32)
    {

      l_type = xl_type_new (XL_TYPE_TYPE_F32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_unref (l_type);

      x_access (xl_data_get_raw (l_data), 0, xf32) = 
	x_access (xl_data_get_raw (l_left->data), 0, xf32) /
	x_access (xl_data_get_raw (l_right->data), 0, xf32);
    }

  x_unref (l_left);
  x_unref (l_right);
  x_unref (p_expr->data);
  p_expr->data = l_data;
}

xvoid
xl_builtin_import (XLExpr* p_expr)
{
  /*
    XLLangBasic* l_lang;
    XLExpr*      l_expr;
    XLData*      l_data;
    XString*     l_text;
    
    l_expr = XL_EXPR (x_list_get (p_expr->list_expr, 0));
    xl_expr_evaluate (l_expr);
    
    l_data = l_expr->data;
    x_unref (l_expr);
    
    if (l_data->type->type == XL_TYPE_TYPE_STRING)
    {
    l_lang = xl_lang_basic_new ();
    l_text = xl_data_get_string (l_data);
    
    l_expr = xl_lang_basic_load (l_lang, l_text);
    if (l_expr)
    {
    xl_expr_evaluate (l_expr);
    x_unref (l_expr);
    }
    
    x_unref (l_lang);
    x_unref (l_text);
    }
  */
}
