/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <xldata.h>
#include <xlclass.h>
#include <xlfield.h>

XLData*
xl_data_new (XLType* p_type, xptr p_ref)
{
  XLData*  self;
  
  self = x_new (XLData);
  x_object_init_object (X_OBJECT (self), xl_data_destroy);
  
  self->type = x_addref (XLType, p_type);
  
  if (p_ref)
    {
      self->ref = TRUE;
      self->raw = p_ref;
      return self;
    }

  self->ref = FALSE;
  self->raw = x_alloc (xl_type_get_size (self->type));
  xl_data_init_data (self);

  return self;
}

xvoid
xl_data_destroy (XObject* self)
{
  return;

  if (XL_DATA (self)->type->array_type == XL_TYPE_ARRAY_TYPE_NONE)
    {
      if (XL_DATA (self)->type->type == XL_TYPE_TYPE_STRING)
	{
	  x_destroy (x_access (XL_DATA (self)->raw, 4, xptr));
	  x_destroy (XL_DATA (self)->raw);
	}
      else
	{
	  x_destroy (XL_DATA (self)->raw);
	}

      x_unref (XL_DATA (self)->type);
    }

  x_destroy (self);
}

xvoid
xl_data_init_data (XLData* self)
{
  XLClass* l_class;
  XLData*  l_data;
  XLField* l_field;
  xu32	   i, l_size;    
  xptr     l_ptr;
  xu32     l_array;

  if (self->type->array_type == XL_TYPE_ARRAY_TYPE_NONE)
    l_array = 1;
  else
    l_array = self->type->array_size;


  if (self->type->type == XL_TYPE_TYPE_STRING)
    {
      self->raw = x_alloc (sizeof (xu32) + sizeof (xptr));
      x_access (self->raw, 0, xu32) = 0;
      l_ptr = x_alloc (1);
      x_access (self->raw, 4, xptr) = l_ptr;
      x_access (l_ptr, 0, xs8) = '\0';
    }

  if (self->type->type == XL_TYPE_TYPE_U32)
    {
      self->raw = x_alloc (sizeof (xu32));
      x_access (self->raw, 0, xu32) = 0;
    }

  if (self->type->type == XL_TYPE_TYPE_F32)
    {
      self->raw = x_alloc (sizeof (xf32));
      x_access (self->raw, 0, xf32) = 0.0;
    }

  if (self->type->type == XL_TYPE_TYPE_POINTER)
    {
      self->raw = x_alloc (sizeof (xptr) * l_array);
      for (i=0; i<l_array; i++)
	x_access (self->raw, i*sizeof(xptr), xptr) = NULL;
    }

  if (self->type->type == XL_TYPE_TYPE_BOOL)
    {
      self->raw = x_alloc (sizeof (xbool));
      x_access (self->raw, 0, xbool) = FALSE;
    }

  if (self->type->type == XL_TYPE_TYPE_STATIC)
    {
      /* We loop in all fields
       */

      l_class = xl_type_get_class (self->type);
      l_size = x_list_get_size (l_class->list_fields);      
      for (i = 0; i < l_size; i++)
	{
	  l_field = XL_FIELD (x_list_get (l_class->list_fields, i));
	  l_data = xl_data_get_field (self, l_field->name);
	  xl_data_init_data (l_data);
	  
	  x_unref (l_field);
	}
    }

  if (self->type->type == XL_TYPE_TYPE_DYNAMIC)
    {
      /* We loop in all fields
       */

      l_class = xl_type_get_class (self->type);
      l_size = x_list_get_size (l_class->list_fields);      
      for (i = 0; i < l_size; i++)
	{
	  l_field = XL_FIELD (x_list_get (l_class->list_fields, i));
	  l_data = xl_data_get_field (self, l_field->name);
	  xl_data_init_data (l_data);
	  
	  x_unref (l_field);
	}      
    }
}

xvoid
xl_data_set_type (XLData* self, XLType* p_type)
{
}

XString*
xl_data_get_string (XLData* self)
{
  XString* l_text;

  l_text = x_string_new ();
  x_string_set_str (l_text, x_access (self->raw, 4, xstr));

  return l_text;
}

xvoid
xl_data_set_string (XLData* self, XString* p_value)
{
  xu32 l_size;

  l_size = x_string_get_size (p_value);

  x_resize (x_access (self->raw, 4, xstr), l_size + 1);
  sprintf (x_access (self->raw, 4, xstr), "%s", x_string_get_str (p_value));
}

xptr
xl_data_get_raw (XLData* self)
{
  return self->raw;
}

xvoid
xl_data_assign (XLData* self, XLData* p_src)
{
  XString* l_text;

  if (!p_src)
    {
      printf ("ASSIGN ERROR: Source not available\n");
      return;
    }

  if (self->type->type == XL_TYPE_TYPE_STRING)
    {
      l_text = xl_data_get_string (p_src);
      xl_data_set_string (self, l_text);
      x_unref (l_text);
    }

  if (self->type->type == XL_TYPE_TYPE_U32)
    {
      x_access (self->raw, 0, xu32) = x_access (p_src->raw, 0, xu32);
    }

  if (self->type->type == XL_TYPE_TYPE_F32)
    {
      x_access (self->raw, 0, xf32) = x_access (p_src->raw, 0, xf32);
    }

  if (self->type->type == XL_TYPE_TYPE_POINTER)
    {
      x_access (self->raw, 0, xptr) = x_access (p_src->raw, 0, xptr);
    }
}

XLData*
xl_data_copy (XLData* self)
{
  return self;
}

XLData*
xl_data_addref_field (XLData* self, XLField* p_field)
{
  return NULL;
}

XLData*
xl_data_convert (XLData* self, XLType* p_type)
{
  return NULL;
}

XLData*
xl_data_get_index (XLData* self, xu32 p_index)
{
  XLData* l_data;
  XLType* l_type;

  l_type = xl_type_new (self->type->type, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
  l_data = xl_data_new (l_type, &x_access (self->raw, (p_index*sizeof(xptr)), xptr));

  return l_data;
}

XLData*
xl_data_get_field (XLData* self, XString* p_fieldname)
{
  XLType*  l_type;
  XLData*  l_data;
  XLField* l_field;

  l_field = xl_class_get_field (xl_type_get_class (self->type), p_fieldname);

  l_type = l_field->type;
  l_data = xl_data_new (l_type, ((xu8*) self->raw) + l_field->index);

  return l_data;
}

xptr
xl_data_get_ptr (XLData* self)
{
  return NULL;
}

