/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <xlfnct.h>
#include <xlmain.h>

xptr (* xl_fnct_direct_void) ();
xptr (* xl_fnct_direct) (int x, ...);

xf32 (* xl_fnct_direct_void_f32) ();
xf32 (* xl_fnct_direct_f32) (int x, ...);

XLData* g_xl_fnct_return = NULL;

XLFnct*
xl_fnct_new ()
{
  XLFnct *self;
  
  self = x_new (XLFnct);
  x_object_init_object (X_OBJECT (self), xl_fnct_destroy);
  
  self->type = XL_FNCT_TYPE_NONE;
  self->expr = NULL;
  self->ret_type = NULL;
  self->name = x_string_new ();
  self->list_params = x_list_new ();
  
  return self;
}

xvoid
xl_fnct_destroy (XObject* self)
{
  x_unref (XL_FNCT (self)->expr);
  x_unref (XL_FNCT (self)->name);
  x_unref (XL_FNCT (self)->ret_type);
  x_unref (XL_FNCT (self)->list_params);
  
  x_destroy (self);
}

xvoid
xl_fnct_set_builtin (XLFnct* self, XString* p_name, XLFnctBuiltin p_builtin)
{
  self->type = XL_FNCT_TYPE_BUILTIN;
  self->builtin = p_builtin;

  x_unref (self->name);
  self->name = x_addref (XString, p_name);
}

xvoid
xl_fnct_set_native (XLFnct* self, XString* p_name, xptr p_native)
{
  self->type = XL_FNCT_TYPE_NATIVE;
  self->native = p_native;

  x_unref (self->name);
  self->name = x_addref (XString, p_name);
}

xvoid
xl_fnct_set_expr (XLFnct* self, XString* p_name, XLExpr* p_expr)
{
  self->type = XL_FNCT_TYPE_EXPR;

  x_unref (self->name);
  self->name = x_addref (XString, p_name);

  x_unref (self->expr);
  self->expr = x_addref (XLExpr, p_expr);
}

xptr
xl_fnct_get_ptr (XLFnct* self)
{
  return NULL;
}

xvoid
xl_fnct_add_param (XLFnct* self, XLField* p_param)
{
  x_list_add (self->list_params, (XObject*) p_param);
}

xvoid
xl_fnct_call_init (XLFnct* self)
{
}

xvoid
xl_fnct_push (XLFnct* self, XLData* p_param)
{
}

XLData*
xl_fnct_call (XLFnct* self, XLExpr* p_expr)
{
  XList*   l_list;
  XLVar*   l_var;
  XLData*  l_data;
  XLData*  l_return;
  XLExpr*  l_expr;
  XLField* l_field;
  XString* l_name;
  xu32     l_size;
  xu32     i, j;
  xptr     l_ret;
  xf32     l_retf;
  xu32     params[100];

  switch (self->type)
    {
    case XL_FNCT_TYPE_BUILTIN:
      (*(self->builtin)) (p_expr);
      return p_expr->data;
      break;
    case XL_FNCT_TYPE_NATIVE:
      j = 0;
      l_size = x_list_get_size (self->list_params);
      for (i=0; i<l_size; i++)
	{
	  l_field = (XLField*) x_list_get (self->list_params, i);
	  l_expr =  XL_EXPR (x_list_get (p_expr->list_expr, i));
	  xl_expr_evaluate (l_expr);

	  if (l_expr->data)
	    {
	      if (l_expr->data->type->type == XL_TYPE_TYPE_STRING)
		params[j] = (xu32) xl_data_get_string (l_expr->data);
	      else
		params[j] = x_access (xl_data_get_raw (l_expr->data), 0, xu32);
	    }

	  j++;
	  
	  x_unref (l_field);
	  x_unref (l_expr);
	}     

      xl_fnct_direct = self->native;
      xl_fnct_direct_void = self->native;
      xl_fnct_direct_f32 = self->native;
      xl_fnct_direct_void_f32 = self->native;

      l_ret = 0;

      if (self->ret_type->type == XL_TYPE_TYPE_F32)
	{
	  switch (j)
	    {
	    case 0:
	      l_retf = (*xl_fnct_direct_void_f32) ();
	      break;
	    case 1:
	      l_retf = xl_fnct_direct_f32 (params[0]);
	      break;
	    case 2:
	      l_retf = (*xl_fnct_direct_f32) (params[0], params[1]);
	      break;
	    case 3:
	      l_retf = (*xl_fnct_direct_f32) (params[0], params[1], params[2]);
	      break;
	    case 4:
	      l_retf = (*xl_fnct_direct_f32) (params[0], params[1], params[2], params[3]);
	      break;
	    case 5:
	      l_retf = (*xl_fnct_direct_f32) (params[0], params[1], params[2], params[3], params[4]);
	      break;
	    case 6:
	      l_retf = (*xl_fnct_direct_f32) (params[0], params[1], params[2], params[3], params[4], params[5]);
	      break;
	    }
	}
      else
	{
	  switch (j)
	    {
	    case 0:
	      l_ret = (*xl_fnct_direct_void) ();
	      break;
	    case 1:
	      l_ret = xl_fnct_direct (params[0]);
	      break;
	    case 2:
	      l_ret = (*xl_fnct_direct) (params[0], params[1]);
	      break;
	    case 3:
	      l_ret = (*xl_fnct_direct) (params[0], params[1], params[2]);
	      break;
	    case 4:
	      l_ret = (*xl_fnct_direct) (params[0], params[1], params[2], params[3]);
	      break;
	    case 5:
	      l_ret = (*xl_fnct_direct) (params[0], params[1], params[2], params[3], params[4]);
	      break;
	    case 6:
	      l_ret = (*xl_fnct_direct) (params[0], params[1], params[2], params[3], params[4], params[5]);
	      break;
	    }
	}

      l_data = xl_data_new (self->ret_type, NULL);
      if (self->ret_type->type == XL_TYPE_TYPE_STRING)
	xl_data_set_string (l_data, (XString*) l_ret);
      else
	{
	  if (self->ret_type->type == XL_TYPE_TYPE_F32)
	    x_access (xl_data_get_raw (l_data), 0, xf32) = l_retf;
	  else
	    x_access (xl_data_get_raw (l_data), 0, xptr) = l_ret;
	}

      return l_data;

      break;
    case XL_FNCT_TYPE_EXPR:

      l_list = xl_main_fnct_enter (g_xl_main_global); 
      if (p_expr)
	{
	  l_size = x_list_get_size (self->list_params);

	  for (i=0; i<l_size; i++)
	    {
	      l_field = (XLField*) x_list_get (self->list_params, i);
	      l_expr =  XL_EXPR (x_list_get (p_expr->list_expr, i));

	      l_name = l_field->name;
	      l_var = xl_var_new (l_name, l_expr->data);
	  
	      xl_main_add_var (g_xl_main_global, l_var, FALSE);

	      x_unref (l_field);
	      x_unref (l_expr);
	      x_unref (l_var);
	    }
	}
      xl_expr_evaluate (self->expr);
      xl_main_fnct_leave (g_xl_main_global, l_list);

      l_return = g_xl_fnct_return;
      g_xl_fnct_return = NULL;
      return l_return;
      break;
    default:
      break;	
    }

  return NULL;
}

