/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <xllangbasic.h>
#include <xllang.h>
#include <xlmain.h>

#define XL_OPER_NONE	0
#define XL_OPER_ADD	1
#define XL_OPER_MIN	2
#define XL_OPER_MUL	3
#define XL_OPER_DIV	4

#define XL_OPER_GE	5
#define XL_OPER_GT	6
#define XL_OPER_LE	7
#define XL_OPER_LT	8
#define XL_OPER_EQ	9
#define XL_OPER_NE	10

/* Prototype
 */

XLExpr*
xl_lang_basic_parse_code (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_code_call (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_decl (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_decl_global (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_decl_function (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_decl_class (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_decl_const (XLLangBasic* self);

XLExpr*
xl_lang_basic_parse_expr (XLLangBasic* self);

XLLangBasic*
xl_lang_basic_new ()
{
  XLLangBasic*	self;
  
  self = x_new (XLLangBasic);
  x_object_init_object (X_OBJECT (self), xl_lang_basic_destroy);

  self->parent.load = xl_lang_basic_load;
  self->idx = 0;
  self->data = NULL;
  self->fileid = NULL;
  self->line = 1;
  self->oper = FALSE;
  
  return self;
}

xvoid
xl_lang_basic_destroy (XObject* self)
{
  x_destroy (self);
}

/* Parser API (tools)
 */

xu32
xl_lang_basic_get_mark (XLLangBasic* self)
{
  return self->idx;
}

xvoid
xl_lang_basic_set_mark (XLLangBasic* self, xu32 p_idx)
{
  self->idx = p_idx;
}

xs8
xl_lang_basic_getc (XLLangBasic* self)
{
  return self->data[self->idx++];
}

xvoid
xl_lang_basic_ungetc (XLLangBasic* self, xs8 p_char)
{
  self->idx--;
}


XString*
xl_lang_basic_get_line (XLLangBasic* self)
{
  return NULL;
}

xbool
xl_lang_basic_check (XLLangBasic* self, xstr p_const)
{
  xl_lang_basic_parse_comments (self);
  
  if (strncmp (p_const, &(self->data[self->idx]), strlen (p_const)) == 0)
    {
      self->idx += strlen (p_const);
      return TRUE;
    }
  else
    return FALSE;
}

xbool
xl_lang_basic_load_file (XLLangBasic* self, XString* p_filename)
{
  xu32	l_size;
  
  self->fileid = fopen(x_string_get_str (p_filename), "rb");
  
  fseek (self->fileid, 0, SEEK_END);
  l_size = ftell (self->fileid) + 1;
  self->data = x_mem_new (l_size);
  fseek (self->fileid, 0, SEEK_SET);
  
  fread (self->data, l_size, 1, self->fileid);
  
  return TRUE;
}

/* Expression parser
 */

xvoid
xl_lang_basic_parse_comments (XLLangBasic* self)
{
  xs8 l_char;
  
  l_char = xl_lang_basic_getc (self);
  while (l_char == ' ' || l_char == '\t' || l_char == '\r' || l_char == '\n')
    l_char = xl_lang_basic_getc (self);
  
  if (l_char == '#')
    {
      l_char = xl_lang_basic_getc (self);
      while (l_char != '\r' && l_char != '\n')
	l_char = xl_lang_basic_getc (self);
      
      xl_lang_basic_ungetc (self, l_char);
      xl_lang_basic_parse_comments (self);
    }
  else	
    xl_lang_basic_ungetc (self, l_char);
}

XString*
xl_lang_basic_parse_name (XLLangBasic* self)
{
  XString*   l_name;
  xs8	     l_char;
  
  xl_lang_basic_parse_comments (self);
  
  l_name = x_string_new ();
  x_string_set_str (l_name, "");
  
  l_char = xl_lang_basic_getc (self);
  if ((l_char >= 'a' && l_char <= 'z') || (l_char >= 'A' && l_char <= 'Z') || 
      l_char == '_' || l_char == '.' || l_char == ':')
    {
      x_string_add_char (l_name, l_char);
      l_char = xl_lang_basic_getc (self);
      while ((l_char >= 'a' && l_char <= 'z') || 
	     (l_char >= 'A' && l_char <= 'Z') || 
	     l_char == '_' || l_char == '.' || l_char == ':' ||
	     (l_char >= '0' && l_char <= '9'))
	{
	  x_string_add_char (l_name, l_char);
	  l_char = xl_lang_basic_getc (self);
	}
    }
  
  xl_lang_basic_ungetc (self, l_char);
  
  return l_name;
}

XLExpr*
xl_lang_basic_parse_type (XLLangBasic* self)
{
  XLExpr*	l_expr;
  XLType*	l_type;
  XString*	l_name;
  XString*	l_class = NULL;
  xs8		ch;
  
  l_name = xl_lang_basic_parse_name (self);
  
  if (x_string_cmp_str (l_name, "") != 0)
    {
      /* Class
       */
      
      ch = xl_lang_basic_getc (self);
      if (ch == '<')
	{
	  l_class = xl_lang_basic_parse_name (self);
	  ch = xl_lang_basic_getc (self);
	}
      else
	xl_lang_basic_ungetc (self, ch);
      
      
      /* Array
       */
      
      ch = xl_lang_basic_getc (self);
      if (ch == '[')
	ch = xl_lang_basic_getc (self);
      else
	xl_lang_basic_ungetc (self, ch);
    }
  
  if (x_string_cmp_str (l_name, "Void") == 0)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_VOID, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);
      
      return l_expr;
    }
  
  if (x_string_cmp_str (l_name, "IntU32") == 0)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);
      
      return l_expr;
    }
  
  if (x_string_cmp_str (l_name, "Pointer") == 0)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_POINTER, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);
      
      return l_expr;
    }
  
  if (x_string_cmp_str (l_name, "String") == 0)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_STRING, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);
      
      return l_expr;
    }
  
  if (x_string_cmp_str (l_name, "Dynamic") == 0)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_DYNAMIC, XL_TYPE_ARRAY_TYPE_NONE, 0, l_class);
      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);
      
      return l_expr;
    }
  
  if (x_string_cmp_str (l_name, "Static") == 0)
    {
      l_type = xl_type_new (XL_TYPE_TYPE_STATIC, XL_TYPE_ARRAY_TYPE_NONE, 0, l_class);
      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);
      
      return l_expr;
    }
  
  return NULL;
}

XLExpr*
xl_lang_basic_parse_data_string (XLLangBasic* self)
{
  XLExpr*	l_expr;
  XLType*	l_type;
  XLData*	l_data;
  XString*	l_text;
  xs8		ch;
  xu32		l_u32;

  xl_lang_basic_parse_comments (self);
  
  ch = xl_lang_basic_getc (self);
  if (ch == '"')
    {
      l_text = x_string_new ();
      
      ch = xl_lang_basic_getc (self);
      while (ch != '"')
	{
	  if (ch == '\\')
	    {
	      ch = xl_lang_basic_getc (self);
	      switch (ch)
		{
		case 'r':
		  x_string_add_char (l_text, '\r');
		  break;
		case 'n':
		  x_string_add_char (l_text, '\n');
		  break;
		case 't':
		  x_string_add_char (l_text, '\t');
		  break;
		default:
		  x_string_add_char (l_text, ch);
		  break;
		}
	    }
	  else
	    x_string_add_char (l_text, ch);
	  
	  ch = xl_lang_basic_getc (self);
	}
      
      l_type = xl_type_new (XL_TYPE_TYPE_STRING, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      xl_data_set_string (l_data, l_text);
      
      l_expr = xl_expr_new ();
      xl_expr_set_data (l_expr, l_data);
      
      return l_expr;
    }
  else
    xl_lang_basic_ungetc (self, ch);
  
  ch = xl_lang_basic_getc (self);
  if (ch >= '0' && ch <= '9')
    {
      l_text = x_string_new ();
      
      while (ch >= '0' && ch <= '9')
	{
	  x_string_add_char (l_text, ch);
	  ch = xl_lang_basic_getc (self);
	}
      xl_lang_basic_ungetc (self, ch);
      
      l_u32 = atoi (x_string_get_str (l_text));
      
      l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      x_access (xl_data_get_raw (l_data), 0, xu32) = l_u32;
      
      l_expr = xl_expr_new ();
      xl_expr_set_data (l_expr, l_data);      
      return l_expr;
    }
  else
    xl_lang_basic_ungetc (self, ch);
  
  return NULL;
}

XLExpr*
xl_lang_basic_parse_data (XLLangBasic* self)
{
  XLExpr*	l_expr;
  
  l_expr = xl_lang_basic_parse_data_string (self);
  if (l_expr)
    return l_expr;
  
  return NULL;
}

XLExpr*
xl_lang_basic_parse_var (XLLangBasic* self)
{
  XLExpr*	l_expr;
  XString*	l_objname;
  XString*	l_fieldname;
  XLExpr*	l_obj;
  XLExpr*	l_field;
  XString*	l_name;
  xu32		i;
  
  l_name = xl_lang_basic_parse_name (self);
  if (x_string_cmp_str (l_name, "") != 0)
    {
      for (i = 0; i < x_string_get_size (l_name); i++)
	{
	  if (x_string_get_char (l_name, i) == '.')
	    {
	      l_objname = x_string_extract (l_name, 0, i);
	      l_fieldname = x_string_extract (l_name, i + 1, x_string_get_size (l_name) - i - 1);
	      
	      l_expr = xl_expr_new_fnct (".");
	      l_obj = xl_expr_new_identifier (x_string_get_str (l_objname));
	      l_field = xl_expr_new_identifier (x_string_get_str (l_fieldname));
	      
	      xl_expr_add (l_expr, l_obj);
	      xl_expr_add (l_expr, l_field);
	      
	      return l_expr;
	    }
	}
      
      l_expr = xl_expr_new_identifier (x_string_get_str (l_name));
      return l_expr;
    }
  else
    return NULL;
}

xu32
xl_lang_basic_parse_oper (XLLangBasic* self)
{
  XString*	l_oper;
  xs8		ch;
  
  xl_lang_basic_parse_comments (self);
  l_oper = x_string_new ();
  
  ch = xl_lang_basic_getc (self);
  while (ch == '+' || ch == '-' || ch == '*' || ch == '/' ||
	 ch == '>' || ch == '<' || ch == '=')
    {
      x_string_add_char (l_oper, ch);
      ch = xl_lang_basic_getc (self);
    }
  xl_lang_basic_ungetc (self, ch);
  
  if (x_string_cmp_str (l_oper, "+") == 0)
    return XL_OPER_ADD;
  
  if (x_string_cmp_str (l_oper, "-") == 0)
    return XL_OPER_MIN;
  
  if (x_string_cmp_str (l_oper, "*") == 0)
    return XL_OPER_MUL;
  
  if (x_string_cmp_str (l_oper, "/") == 0)
    return XL_OPER_DIV;
  
  if (x_string_cmp_str (l_oper, ">") == 0)
    return XL_OPER_GT;
  
  if (x_string_cmp_str (l_oper, "<") == 0)
    return XL_OPER_LT;
  
  if (x_string_cmp_str (l_oper, ">=") == 0)
    return XL_OPER_GE;
  
  if (x_string_cmp_str (l_oper, "<=") == 0)
    return XL_OPER_LE;
  
  if (x_string_cmp_str (l_oper, "==") == 0)
    return XL_OPER_EQ;
  
  if (x_string_cmp_str (l_oper, "<>") == 0)
    return XL_OPER_NE;
  
  
  return XL_OPER_NONE;
}

XLExpr*
xl_lang_basic_parse_expr_oper (XLLangBasic* self)
{
  XLExpr*	l_expr;
  XLExpr*	l_expr1;
  XLExpr*	l_expr2;
  xu32		l_mark;
  xbool		l_oper;
  xstr		l_op;

  self->oper = TRUE;
  
  l_mark = xl_lang_basic_get_mark (self);
  
  l_expr1 = xl_lang_basic_parse_expr (self);
  if (l_expr1)
    {
      switch (xl_lang_basic_parse_oper (self))
	{
	case XL_OPER_ADD:
	  l_op = "+";
	  break;
	case XL_OPER_MIN:
	  l_op = "-";
	  break;
	case XL_OPER_MUL:
	  l_op = "*";
	  break;
	case XL_OPER_DIV:
	  l_op = "/";
	  break;
	case XL_OPER_GT:
	  l_op = ">";
	  break;
	case XL_OPER_GE:
	  l_op = ">=";
	  break;
	case XL_OPER_LT:
	  l_op = "<";
	  break;
	case XL_OPER_LE:
	  l_op = "<=";
	  break;
	case XL_OPER_NE:
	  l_op = "<>";
	  break;
	case XL_OPER_EQ:
	  l_op = "==";
	  break;	
	default:
	  xl_lang_basic_set_mark (self, l_mark);
	  self->oper = FALSE;
	  return NULL;
	}
      l_mark = xl_lang_basic_get_mark (self);
      l_expr2 = xl_lang_basic_parse_expr (self);
      l_oper = xl_lang_basic_parse_oper (self);
      xl_lang_basic_set_mark (self, l_mark);		
      
      if (l_oper)
	l_expr2 = xl_lang_basic_parse_expr_oper (self);
      else
	l_expr2 = xl_lang_basic_parse_expr (self);
      
      l_expr = xl_expr_new_fnct (l_op);
      xl_expr_add (l_expr, l_expr1);
      xl_expr_add (l_expr, l_expr2);
      
      self->oper = FALSE;
      return l_expr;
      
    }
  
  xl_lang_basic_set_mark (self, l_mark);
  self->oper = FALSE;
  return NULL;
}

XLExpr*
xl_lang_basic_parse_code_if (XLLangBasic* self)
{
  XLExpr*  l_expr;
  XLExpr*  l_cond;
  XLExpr*  l_codetrue;
  XLExpr*  l_codefalse;
  xu32	   l_mark;

  l_mark = xl_lang_basic_get_mark (self);
  
  if (xl_lang_basic_check (self, "If"))
    {
      xl_lang_basic_check (self, "(");
      l_cond = xl_lang_basic_parse_expr (self);
      xl_lang_basic_check (self, ")");
      
      xl_lang_basic_check (self, "Then");
      
      l_codetrue = xl_lang_basic_parse_code (self);
      
      if (xl_lang_basic_check (self, "Else"))
	l_codefalse = xl_lang_basic_parse_code (self);
      else
	l_codefalse = NULL;
      
      xl_lang_basic_check (self, "End If");
      
      l_expr = xl_expr_new_fnct ("if");
      xl_expr_add (l_expr, l_cond);
      xl_expr_add (l_expr, l_codetrue);
      if (l_codefalse)
	xl_expr_add (l_expr, l_codefalse);
      
      return l_expr;
    }
  
  xl_lang_basic_set_mark (self, l_mark);
  return NULL;
}

XLExpr*
xl_lang_basic_parse_const (XLLangBasic* self)
{
  XLExpr*	l_expr;
  XLType*	l_type;
  XLData*	l_data;
  XString*	l_name;
  xu32		l_mark;

  l_mark = xl_lang_basic_get_mark (self);
  
  l_name = xl_lang_basic_parse_name (self);
  if (x_string_cmp_str (l_name, "") != 0)
    {
      if (x_string_cmp_str (l_name, "True") == 0)
		{
		  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
		  l_data = xl_data_new (l_type, NULL);
		  x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
		  
		  l_expr = xl_expr_new ();
		  xl_expr_set_data (l_expr, l_data);      
		  
		  return l_expr;
		}
      
      if (x_string_cmp_str (l_name, "False") == 0)
	{
	  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
	  l_data = xl_data_new (l_type, NULL);
	  x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;
	  
	  l_expr = xl_expr_new ();
	  xl_expr_set_data (l_expr, l_data);      
	  
	  return l_expr;
	}
    }
  
  xl_lang_basic_set_mark (self, l_mark);
  return NULL;
}


XLExpr*
xl_lang_basic_parse_expr (XLLangBasic* self)
{
	XLExpr*	l_expr;

	if (!self->oper)
	{
		l_expr = xl_lang_basic_parse_expr_oper (self);
		if (l_expr)
			return l_expr;
	}

	l_expr = xl_lang_basic_parse_const (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_data (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_code_call (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_var (self);
	if (l_expr)
		return l_expr;


	return NULL;
}

XLExpr*
xl_lang_basic_parse_identifier (XLLangBasic* self)
{
	XLExpr*		l_identifier;
	XString*	l_name;

	l_name = xl_lang_basic_parse_name (self);
	if (x_string_cmp_str (l_name, "") != 0)
	{
		l_identifier = xl_expr_new ();
		xl_expr_set_identifier (l_identifier, l_name);

		return l_identifier;
	}
	else
		return NULL;
}

XLExpr*
xl_lang_basic_parse_code_local (XLLangBasic* self)
{
	XLExpr*	l_identifier;
	XLExpr*	l_type;
	XLExpr*	l_expr;
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Local"))
	{
		l_identifier = xl_lang_basic_parse_identifier (self);
		if (!l_identifier)
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}
		if (!xl_lang_basic_check (self, "As"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}
		l_type = xl_lang_basic_parse_type (self);
		
		if (!l_type)
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("local");
		xl_expr_add (l_expr, l_type);
		xl_expr_add (l_expr, l_identifier);

		x_unref (l_type);
		x_unref (l_identifier);
		
		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);

	return NULL;
}

XLExpr*
xl_lang_basic_invoke (XString* l_name)
{
  XLExpr*   l_expr;
  XLExpr*   l_obj;
  XLExpr*   l_meth;
  XString*  l_objname;
  XString*  l_methname;
  xu32	    i;
  
  for (i = 0; i < x_string_get_size (l_name); i++)
    {
      if (x_string_get_char (l_name, i) == '.')
	{
	  l_objname = x_string_extract (l_name, 0, i);
	  l_methname = x_string_extract (l_name, i + 1, x_string_get_size (l_name) - i - 1);
	  
	  l_expr = xl_expr_new_fnct ("::");
	  l_obj = xl_expr_new_identifier (x_string_get_str (l_objname));
	  l_meth = xl_expr_new_identifier (x_string_get_str (l_methname));
	  
	  xl_expr_add (l_expr, l_obj);
	  xl_expr_add (l_expr, l_meth);
	  
	  return l_expr;
	}
    }
  
  l_expr = xl_expr_new_fnct (x_string_get_str (l_name));
  return l_expr;
}

XLExpr*
xl_lang_basic_parse_code_call (XLLangBasic* self)
{
	XLExpr*		l_expr;
	XLExpr*		l_param;
	XString*	l_name;
	xu32		l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	l_name = xl_lang_basic_parse_name (self);
	if (x_string_cmp_str (l_name, "") != 0)
	{
		if (!xl_lang_basic_check (self, "("))
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		/* If NNNN.NNNN then its an invoke instead of
		 * a standard call.
		 */

		l_expr = xl_lang_basic_invoke (l_name);

		l_param = xl_lang_basic_parse_expr (self);
		while (l_param)
		{
			xl_expr_add (l_expr, l_param);

			if (xl_lang_basic_check (self, ","))
				l_param = xl_lang_basic_parse_expr (self);
			else
				l_param = NULL;
		}

		if (!xl_lang_basic_check (self, ")"))
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_code_assign (XLLangBasic* self)
{
	XLExpr*		l_dest;
	XLExpr*		l_src;
	XLExpr*		l_expr;
	xu32		l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	l_dest = xl_lang_basic_parse_var (self);
	if (l_dest)
	{
		if (!xl_lang_basic_check (self, "="))
		{
			x_unref (l_dest);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_src = xl_lang_basic_parse_expr (self);
		if (!l_src)
		{
			x_unref (l_dest);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("=");
		xl_expr_add (l_expr, l_dest);
		xl_expr_add (l_expr, l_src);

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_code_while (XLLangBasic* self)
{
	XLExpr*	l_expr;
	XLExpr*	l_cond;
	XLExpr*	l_code;
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "While"))
	{
		xl_lang_basic_check (self, "(");
		l_cond = xl_lang_basic_parse_expr (self);
		xl_lang_basic_check (self, ")");

		xl_lang_basic_check (self, "Then");

		l_code = xl_lang_basic_parse_code (self);
		
		xl_lang_basic_check (self, "End While");

		l_expr = xl_expr_new_fnct ("while");
		xl_expr_add (l_expr, l_cond);
		xl_expr_add (l_expr, l_code);

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_code_return (XLLangBasic* self)
{
	XLExpr*	l_expr;
	XLExpr*	l_value;
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Return"))
	{
		xl_lang_basic_check (self, "(");
		l_value = xl_lang_basic_parse_expr (self);
		xl_lang_basic_check (self, ")");

		l_expr = xl_expr_new_fnct ("return");
		xl_expr_add (l_expr, l_value);

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_code_line (XLLangBasic* self)
{
	XLExpr*	l_expr;

	l_expr = xl_lang_basic_parse_code_local (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_code_if (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_code_while (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_code_return (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_code_call (self);
	if (l_expr)
		return l_expr;


	l_expr = xl_lang_basic_parse_code_assign (self);
	if (l_expr)
		return l_expr;

	return NULL;
}

XLExpr*
xl_lang_basic_parse_code (XLLangBasic* self)
{
	XLExpr*	l_expr;
	XLExpr*	l_code;

	l_expr = xl_expr_new_fnct ("@");

	l_code = xl_lang_basic_parse_code_line (self);
	while (l_code)
	{
		xl_expr_add (l_expr, l_code);
		l_code = xl_lang_basic_parse_code_line (self);
	}

	return l_expr;
}

/* Top-level parser
 */

XLExpr*
xl_lang_basic_load (XLLang* self, XString* p_filename)
{
  XLExpr*	l_expr;
  XLExpr*	l_subexpr;
  
  l_expr = xl_expr_new ();
  xl_expr_set_block (l_expr);
  
  if (xl_lang_basic_load_file (XL_LANG_BASIC (self), p_filename))
    {
      l_subexpr = xl_lang_basic_parse_decl (XL_LANG_BASIC (self));
      while (l_subexpr)
	{
	  xl_expr_add (l_expr, l_subexpr);
	  l_subexpr = xl_lang_basic_parse_decl (XL_LANG_BASIC (self));
	}
    }
  
  return l_expr;
}

XLExpr*
xl_lang_basic_parse_decl (XLLangBasic* self)
{
	XLExpr*	l_expr;

	xl_lang_basic_parse_comments (self);

	l_expr = xl_lang_basic_parse_decl_global (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_decl_function (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_decl_class (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_decl_const (self);
	if (l_expr)
		return l_expr;
	
	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_global (XLLangBasic* self)
{
	XLExpr*	l_identifier;
	XLExpr*	l_type;
	XLExpr*	l_expr;
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Global"))
	{
		l_identifier = xl_lang_basic_parse_identifier (self);
		if (!l_identifier)
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}
		if (!xl_lang_basic_check (self, "As"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}
		l_type = xl_lang_basic_parse_type (self);
		
		if (!l_type)
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("global");
		xl_expr_add (l_expr, l_type);
		xl_expr_add (l_expr, l_identifier);

		x_unref (l_type);
		x_unref (l_identifier);
		
		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);

	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_function (XLLangBasic* self)
{
	xu32	i;
	xu32	l_mark;
	XList*	l_paramsname;
	XList*	l_paramstype;

	XLExpr*	l_expr;
	XLExpr*	l_code;
	XLExpr*	l_identifier;
	XLExpr* l_rettype;
	XLExpr*	l_paramname;
	XLExpr*	l_paramtype;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Function"))
	{
		l_identifier = xl_lang_basic_parse_identifier (self);
		if (!l_identifier)
		{
			printf ("ERROR ('Function'): Expected identifier\n");
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		if (!xl_lang_basic_check (self, "("))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		/* Parameters
		 */

		l_paramsname = x_list_new ();
		l_paramstype = x_list_new ();

		l_paramname = xl_lang_basic_parse_identifier (self);
		while (l_paramname)
		{
			xl_lang_basic_check (self, "As");
			l_paramtype = xl_lang_basic_parse_type (self);

			x_list_add (l_paramsname, X_OBJECT (l_paramname));
			x_list_add (l_paramstype, X_OBJECT (l_paramtype));
	
			l_paramname = xl_lang_basic_parse_identifier (self);
		}

		if (!xl_lang_basic_check (self, ")"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		if (!xl_lang_basic_check (self, "As"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_rettype = xl_lang_basic_parse_type (self);
		if (!l_rettype)
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		/* Code
		 */

		l_code = xl_lang_basic_parse_code (self);

		/* End Function
		 */

		if (!xl_lang_basic_check (self, "End Function"))
		{
			printf ("ERROR ('Function'): Expected 'End Function'\n");
			x_unref (l_identifier);
			x_unref (l_rettype);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("fnct");
		xl_expr_add (l_expr, l_rettype);
		xl_expr_add (l_expr, l_identifier);


		for (i = 0; i < x_list_get_size (l_paramsname); i++)
		{
			l_paramname = XL_EXPR (x_list_get (l_paramsname, i));
			l_paramtype = XL_EXPR (x_list_get (l_paramstype, i));

			xl_expr_add (l_expr, l_paramtype);
			xl_expr_add (l_expr, l_paramname);
		}

		xl_expr_add (l_expr, l_code);

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_class_property (XLLangBasic* self)
{
	XLExpr*	l_identifier;
	XLExpr*	l_type;
	XLExpr*	l_expr;
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Property"))
	{
		l_identifier = xl_lang_basic_parse_identifier (self);
		if (!l_identifier)
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}
		if (!xl_lang_basic_check (self, "As"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}
		l_type = xl_lang_basic_parse_type (self);
		
		if (!l_type)
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("prop");
		xl_expr_add (l_expr, l_type);
		xl_expr_add (l_expr, l_identifier);

		x_unref (l_type);
		x_unref (l_identifier);
		
		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_class_method (XLLangBasic* self)
{
	xu32	i;
	xu32	l_mark;
	XList*	l_paramsname;
	XList*	l_paramstype;

	XLExpr*	l_expr;
	XLExpr*	l_code;
	XLExpr*	l_identifier;
	XLExpr* l_rettype;
	XLExpr*	l_paramname;
	XLExpr*	l_paramtype;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Method"))
	{
		l_identifier = xl_lang_basic_parse_identifier (self);
		if (!l_identifier)
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		if (!xl_lang_basic_check (self, "("))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		/* Parameters
		 */

		l_paramsname = x_list_new ();
		l_paramstype = x_list_new ();

		l_paramname = xl_lang_basic_parse_identifier (self);
		while (l_paramname)
		{
			xl_lang_basic_check (self, "As");
			l_paramtype = xl_lang_basic_parse_type (self);

			x_list_add (l_paramsname, X_OBJECT (l_paramname));
			x_list_add (l_paramstype, X_OBJECT (l_paramtype));
	
			l_paramname = xl_lang_basic_parse_identifier (self);
		}

		if (!xl_lang_basic_check (self, ")"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		if (!xl_lang_basic_check (self, "As"))
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_rettype = xl_lang_basic_parse_type (self);
		if (!l_rettype)
		{
			x_unref (l_identifier);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		/* Code
		 */

		l_code = xl_lang_basic_parse_code (self);

		/* End Function
		 */

		if (!xl_lang_basic_check (self, "End Method"))
		{
			x_unref (l_identifier);
			x_unref (l_rettype);
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("meth");
		xl_expr_add (l_expr, l_rettype);
		xl_expr_add (l_expr, l_identifier);


		for (i = 0; i < x_list_get_size (l_paramsname); i++)
		{
			l_paramname = XL_EXPR (x_list_get (l_paramsname, i));
			l_paramtype = XL_EXPR (x_list_get (l_paramstype, i));

			xl_expr_add (l_expr, l_paramtype);
			xl_expr_add (l_expr, l_paramname);
		}

		xl_expr_add (l_expr, l_code);

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_class_member (XLLangBasic* self)
{
	XLExpr*	l_expr;

	l_expr = xl_lang_basic_parse_decl_class_property (self);
	if (l_expr)
		return l_expr;

	l_expr = xl_lang_basic_parse_decl_class_method (self);
	if (l_expr)
		return l_expr;

	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_class (XLLangBasic* self)
{
	XLExpr*	l_expr;
	XLExpr*	l_member;
	XLExpr*	l_identifier;
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Class"))
	{
		l_identifier = xl_lang_basic_parse_identifier (self);
		if (!l_identifier)
		{
			xl_lang_basic_set_mark (self, l_mark);
			return NULL;
		}

		l_expr = xl_expr_new_fnct ("class");
		xl_expr_add (l_expr, l_identifier);

		/* Members
		 */ 
		l_member = xl_lang_basic_parse_decl_class_member (self);
		while (l_member)
		{
			xl_expr_add (l_expr, l_member);
			l_member = xl_lang_basic_parse_decl_class_member (self);
		}

		if (!xl_lang_basic_check (self, "End Class"))
		{
			return NULL;	
		}

		return l_expr;
	}

	xl_lang_basic_set_mark (self, l_mark);
	return NULL;
}

XLExpr*
xl_lang_basic_parse_decl_const (XLLangBasic* self)
{
	xu32	l_mark;

	l_mark = xl_lang_basic_get_mark (self);

	if (xl_lang_basic_check (self, "Const"))
	{
		printf ("Const detected\n");
		return xl_expr_new ();
	}

	xl_lang_basic_set_mark (self, l_mark);

	return NULL;
}



