/* X/lang - the eXtensible Language
 * Copyright (C) 2001 Patrick Deschenes
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* These files are distributed at http://xlang.sourceforge.net
 */

#include <xllanglisp.h>
#include <xlmain.h>

XLLangLisp*
xl_lang_lisp_new ()
{
  XLLangLisp *self;
  
  self = x_new (XLLangLisp);
  x_object_init_object (X_OBJECT (self), xl_lang_lisp_destroy);
  
  self->parent.load = xl_lang_lisp_load;
  self->fileid = NULL;
  self->line = 1;
  
  return self;
}

xvoid
xl_lang_lisp_destroy (XObject* self)
{
  x_destroy (self);
}

xs8
xl_lang_lisp_getc (XLLangLisp* self)
{
  xs8 c;
  c = getc (self->fileid);
  if (c == '\n')
    self->line++;
  
  return c;
}

xvoid
xl_lang_lisp_ungetc (XLLangLisp* self, xs8 p_char)
{
  if (p_char == '\n')
    self->line--;

  ungetc (p_char, self->fileid);
}

XLExpr*
xl_lang_lisp_load (XLLang* self, XString* p_filename)
{
  XLExpr* l_expr;
  XLExpr* l_subexpr;

  XL_LANG_LISP (self)->fileid = fopen (x_string_get_str (p_filename), "rt");  

  if (XL_LANG_LISP (self)->fileid)
    {
      l_expr = xl_expr_new ();
      xl_expr_set_block (l_expr);
      
      l_subexpr = xl_lang_lisp_parse (XL_LANG_LISP (self));
      while (l_subexpr)
	{
	  xl_expr_add (l_expr, l_subexpr);
	  x_unref (l_subexpr);
	  l_subexpr = xl_lang_lisp_parse (XL_LANG_LISP (self));
	}
      
      fclose (XL_LANG_LISP (self)->fileid);

      return l_expr;	
    }
  else
    {
      return NULL;
    }
}

XLExpr*
xl_lang_lisp_parse (XLLangLisp* self)
{
  XLExpr*  l_expr;
  XLExpr*  l_subexpr;
  XString* l_name;
  xs8	   l_char;

  xl_lang_lisp_parse_comments (self);
  
  l_char = xl_lang_lisp_getc (self);
  if (l_char == '(')
    {
      l_name = xl_lang_lisp_parse_name (self);
     
      l_expr = xl_expr_new ();
      xl_expr_set_fnct (l_expr, l_name);
      x_unref (l_name);
      
      l_subexpr = xl_lang_lisp_parse (self);
      while (l_subexpr)
	{
	  xl_expr_add (l_expr, l_subexpr);
	  x_unref (l_subexpr);
	  l_subexpr = xl_lang_lisp_parse (self);
	}
      
      xl_lang_lisp_parse_comments (self);	
      l_char = xl_lang_lisp_getc (self);
      if (l_char == ')')
	{
	  return l_expr;
	}
      else
	{
	  printf ("ERROR: Unterminated expression at line %d\n", self->line);
	  return NULL;
	}
    }
  else
    {	
      xl_lang_lisp_ungetc (self, l_char);

      l_expr = xl_lang_lisp_parse_data (self);
      if (l_expr)
	return l_expr;
   
      l_expr = xl_lang_lisp_parse_fnct (self);
      if (l_expr)
	return l_expr;
         
      l_expr = xl_lang_lisp_parse_type (self);
      if (l_expr)
	return l_expr;
      
      l_expr = xl_lang_lisp_parse_identifier (self);
      if (l_expr)
	return l_expr;
    
      return NULL;
    }
}

xvoid
xl_lang_lisp_parse_comments (XLLangLisp* self)
{
  xs8 l_char;
  
  l_char = xl_lang_lisp_getc (self);
  while (l_char == ' ' || l_char == '\t' || l_char == '\r' || l_char == '\n')
    l_char = xl_lang_lisp_getc (self);
  
  if (l_char == '#')
    {
      l_char = xl_lang_lisp_getc (self);
      while (l_char != '\r' && l_char != '\n')
	l_char = xl_lang_lisp_getc (self);
      
      xl_lang_lisp_ungetc (self, l_char);
      xl_lang_lisp_parse_comments (self);
    }
  else	
    xl_lang_lisp_ungetc (self, l_char);
}

XString*
xl_lang_lisp_parse_name (XLLangLisp* self)
{
  XString* l_name;
  xs8	   l_char;
  
  xl_lang_lisp_parse_comments (self);
  
  l_name = x_string_new ();
  x_string_set_str (l_name, "");
  
  l_char = xl_lang_lisp_getc (self);
  if ((l_char >= 'a' && l_char <= 'z') || (l_char >= 'A' && l_char <= 'Z') || 
      l_char == '@' || l_char == '_' || l_char == '=' || l_char == '<' ||
      l_char == '>' || l_char == '+' || l_char == '-' || l_char == '*' ||
      l_char == '/' || l_char == '.' || l_char == ':' || l_char == '!')
    {
      x_string_add_char (l_name, l_char);
      l_char = xl_lang_lisp_getc (self);
      while ((l_char >= 'a' && l_char <= 'z') || 
	     (l_char >= 'A' && l_char <= 'Z') || 
	     l_char == '@' || l_char == '_' || 
	     l_char == '=' || l_char == '<' ||
	     l_char == '>' || l_char == '+' ||
	     l_char == '-' || l_char == '*' ||
	     l_char == '/' || l_char == '.' ||
	     l_char == ':' || l_char == '!' ||
	     (l_char >= '0' && l_char <= '9'))
	{
	  x_string_add_char (l_name, l_char);
	  l_char = xl_lang_lisp_getc (self);
	}
    }
  
  xl_lang_lisp_ungetc (self, l_char);
  
  return l_name;
}

XLExpr*
xl_lang_lisp_parse_type (XLLangLisp* self)
{
  XLExpr*  l_expr;
  XLType*  l_type;
  XString* l_name;
  xs8	   l_char;
  
  xl_lang_lisp_parse_comments (self);
  
  l_char = xl_lang_lisp_getc (self);
  if (l_char == '%')
    {
      l_name = xl_lang_lisp_parse_name (self);
      
      l_char = xl_lang_lisp_getc (self);
      if (l_char == '[')
	{
	  l_char = xl_lang_lisp_getc (self);
	  if (l_char ==  ']')
	    {					
	    }
	  else
	    {
	      printf ("ERROR: Incomplete array type []\n");
	    }
	}
      else
	xl_lang_lisp_ungetc (self, l_char);

      l_type = NULL;
      if (x_string_cmp_str (l_name, "string") == 0)
	l_type = xl_type_new (XL_TYPE_TYPE_STRING, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);

      if (x_string_cmp_str (l_name, "u32") == 0)
	l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);

      if (x_string_cmp_str (l_name, "pointer") == 0)
	l_type = xl_type_new (XL_TYPE_TYPE_POINTER, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);

      if (x_string_cmp_str (l_name, "void") == 0)
	l_type = xl_type_new (XL_TYPE_TYPE_VOID, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);

      if (x_string_cmp_str (l_name, "bool") == 0)
	l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);

      if (x_string_cmp_str (l_name, "static") == 0)
	{
	  l_char = xl_lang_lisp_getc (self);
	  l_name = xl_lang_lisp_parse_name (self);
	  l_char = xl_lang_lisp_getc (self);

	  l_type = xl_type_new (XL_TYPE_TYPE_STATIC, XL_TYPE_ARRAY_TYPE_NONE, 0, l_name);
	}

      if (x_string_cmp_str (l_name, "dynamic") == 0)
	{
	  l_char = xl_lang_lisp_getc (self);
	  l_name = xl_lang_lisp_parse_name (self);
	  l_char = xl_lang_lisp_getc (self);

	  l_type = xl_type_new (XL_TYPE_TYPE_DYNAMIC, XL_TYPE_ARRAY_TYPE_NONE, 0, l_name);
	}

      l_expr = xl_expr_new ();
      xl_expr_set_type (l_expr, l_type);

      x_unref (l_type);
      x_unref (l_name);

      return l_expr;
    }
  else
    {
      xl_lang_lisp_ungetc (self, l_char);
      return NULL;
    }
}

XLExpr*
xl_lang_lisp_parse_data (XLLangLisp* self)
{
  XLExpr*  l_expr;
  XLData*  l_data;
  XLType*  l_type;
  XString* l_text;
  xs8	   l_char;
  xu32     l_u32;
  
  xl_lang_lisp_parse_comments (self);
  
  l_char = xl_lang_lisp_getc (self);
  if (l_char == '"')
    {	
      l_text = x_string_new ();
      x_string_set_str (l_text, "");
      
      l_char = xl_lang_lisp_getc (self);
      while (l_char != '"')
	{
	  if (l_char == '\\')
	    {
	      l_char = xl_lang_lisp_getc (self);
	      switch (l_char)
		{
		case 'r':
		  l_char = '\r';
		  break;
		case 'n':
		  l_char = '\n';
		  break;
		case 't':
		  l_char = '\t';
		  break;
		default:
		  break;
		}
	    }
	  x_string_add_char (l_text, l_char);
	  
	  l_char = xl_lang_lisp_getc (self);
	}
      
      l_type = xl_type_new (XL_TYPE_TYPE_STRING, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
      l_data = xl_data_new (l_type, NULL);
      xl_data_set_string (l_data, l_text);
      
      l_expr = xl_expr_new ();
      xl_expr_set_data (l_expr, l_data);

      x_unref (l_type);
      x_unref (l_text);
      x_unref (l_data);
      
      return l_expr;
    }
  else
    {
      if (l_char >= '0' && l_char <= '9')
	{
	  l_text = x_string_new ();
	  x_string_set_str (l_text, "");
	  x_string_add_char (l_text, l_char);
	  l_char = xl_lang_lisp_getc (self);
	  while (l_char >= '0' && l_char <= '9')
	    {
	      x_string_add_char (l_text, l_char);
	      l_char = xl_lang_lisp_getc (self);
	    }
	  xl_lang_lisp_ungetc (self, l_char);
	  
	  l_u32 = atoi (x_string_get_str (l_text));

	  l_type = xl_type_new (XL_TYPE_TYPE_U32, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
	  l_data = xl_data_new (l_type, NULL);
	  x_access (xl_data_get_raw (l_data), 0, xu32) = l_u32;

	  l_expr = xl_expr_new ();
	  xl_expr_set_data (l_expr, l_data);      
	  
	  x_unref (l_type);
	  x_unref (l_data);
	  x_unref (l_text);
	  
	  return l_expr;
	}
      else
	{
	  if (l_char == '$')
	    {
	      l_text = xl_lang_lisp_parse_name (self);	      
	      if (x_string_cmp_str (l_text, "true") == 0)
		{
		  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
		  l_data = xl_data_new (l_type, NULL);
		  x_access (xl_data_get_raw (l_data), 0, xbool) = TRUE;
		  
		  l_expr = xl_expr_new ();
		  xl_expr_set_data (l_expr, l_data);      

		  x_unref (l_text);
		  x_unref (l_type);
		  x_unref (l_data);
		  
		  return l_expr;	
		}
	      if (x_string_cmp_str (l_text, "false") == 0)
		{
		  l_type = xl_type_new (XL_TYPE_TYPE_BOOL, XL_TYPE_ARRAY_TYPE_NONE, 0, NULL);
		  l_data = xl_data_new (l_type, NULL);
		  x_access (xl_data_get_raw (l_data), 0, xbool) = FALSE;
		  
		  l_expr = xl_expr_new ();
		  xl_expr_set_data (l_expr, l_data);      

		  x_unref (l_text);
		  x_unref (l_type);
		  x_unref (l_data);
		  
		  return l_expr;	
		}

	      x_unref (l_text);
	    }
	  else
	    xl_lang_lisp_ungetc (self, l_char);
	}

      return NULL;
    }
}

XLExpr*
xl_lang_lisp_parse_fnct (XLLangLisp* self)
{
  return NULL;
}

XLExpr*
xl_lang_lisp_parse_identifier (XLLangLisp* self)
{
  XLExpr*  l_expr;
  XString* l_name;
  
  xl_lang_lisp_parse_comments (self);
  
  l_name = xl_lang_lisp_parse_name (self);
  
  if (x_string_cmp_str (l_name, "") != 0)
    {		
      l_expr = xl_expr_new ();		
      xl_expr_set_identifier (l_expr, l_name);

      x_unref (l_name);
      return l_expr;
    }
  else
    {
      x_unref (l_name);
      return NULL;
    }
}
